"""
Title: SpatialBasePlot for all spatial coordinates and image plot
Author: Duy Pham
Date: 20 Feb 2021
"""

from lib2to3.pgen2.token import OP
from typing import Optional, Union, Mapping, List  # Special
from typing import Sequence, Iterable  # ABCs
from typing import Tuple  # Classes

import numbers
import numpy as np
import pandas as pd
from anndata import AnnData

from matplotlib import rcParams, ticker, gridspec, axes
import matplotlib.pyplot as plt
import matplotlib
from scipy.interpolate import griddata
import networkx as nx

from ..classes import Spatial
from ..utils import _AxesSubplot, Axes, _read_graph
from .utils import centroidpython, get_cluster, get_node, check_sublist, get_cmap

################################################################
#                                                              #
#                  Spatial base plot class                     #
#                                                              #
################################################################


class SpatialBasePlot(Spatial):
    def __init__(
        self,
        # plotting param
        adata: AnnData,
        title: Optional["str"] = None,
        figsize: Optional[Tuple[float, float]] = None,
        cmap: Optional[str] = "Spectral_r",
        use_label: Optional[str] = None,
        list_clusters: Optional[list] = None,
        ax: Optional[matplotlib.axes._subplots.Axes] = None,
        fig: Optional[matplotlib.figure.Figure] = None,
        show_plot: Optional[bool] = True,
        show_axis: Optional[bool] = False,
        show_image: Optional[bool] = True,
        show_color_bar: Optional[bool] = True,
        color_bar_label: Optional[str] = "",
        zoom_coord: Optional[float] = None,
        crop: Optional[bool] = True,
        margin: Optional[bool] = 100,
        size: Optional[float] = 7,
        image_alpha: Optional[float] = 1.0,
        cell_alpha: Optional[float] = 0.7,
        use_raw: Optional[bool] = False,
        fname: Optional[str] = None,
        dpi: Optional[int] = 120,
        **kwds,
    ):
        super().__init__(
            adata,
        )
        self.title = title
        self.figsize = figsize
        self.image_alpha = image_alpha
        self.cell_alpha = cell_alpha
        self.size = size
        self.query_adata = self.adata[0].copy()
        self.list_clusters = list_clusters
        self.fname = fname
        self.dpi = dpi

        if use_raw:
            self.query_adata = self.adata[0].raw.to_adata().copy()

        if self.list_clusters != None:
            assert use_label != None, "Please specify `use_label` parameter!"

        if use_label != None:

            assert (
                use_label in self.adata[0].obs.columns
            ), "Please choose the right label in `adata.obs.columns`!"
            self.use_label = use_label

            if self.list_clusters is None:

                self.list_clusters = np.array(
                    self.adata[0].obs[use_label].cat.categories
                )
            else:
                if type(self.list_clusters) != list:
                    self.list_clusters = [self.list_clusters]

                clusters_indexes = [
                    np.where(adata.obs[use_label].cat.categories == i)[0][0]
                    for i in self.list_clusters
                ]
                self.list_clusters = np.array(self.list_clusters)[
                    np.argsort(clusters_indexes)
                ]

            self.query_indexes = self._get_query_clusters_index()

            self._select_clusters()

        # Initialize cmap
        scanpy_cmap = ["vega_10_scanpy", "vega_20_scanpy", "default_102", "default_28"]
        stlearn_cmap = ["jana_40", "default"]
        cmap_available = plt.colormaps() + scanpy_cmap + stlearn_cmap
        error_msg = (
            "cmap must be a matplotlib.colors.LinearSegmentedColormap OR"
            "one of these: " + str(cmap_available)
        )
        if type(cmap) == str:
            assert cmap in cmap_available, error_msg
        elif type(cmap) != matplotlib.colors.LinearSegmentedColormap:
            raise Exception(error_msg)
        self.cmap = cmap

        if type(fig) == type(None) and type(ax) == type(None):
            self.fig, self.ax = self._generate_frame()
        else:
            self.fig, self.ax = fig, ax

        if show_axis == False:
            self._remove_axis(self.ax)

        if show_image:
            self._add_image(self.ax)

        if zoom_coord is not None:
            self._zoom_image(self.ax, zoom_coord)
            crop = False
        if crop:
            self._crop_image(self.ax, margin)

    def _select_clusters(self):
        def create_query(list_cl, use_label):
            ini = ""
            for sub in list_cl:
                ini = ini + self.use_label + ' == "' + str(sub) + '" | '
            return ini[:-2]

        if self.list_clusters is not None:
            # IF not all clusters specified, subset, otherwise just copy.
            if len(self.list_clusters) != len(
                self.adata[0].obs[self.use_label].cat.categories
            ):
                self.query_adata = self.query_adata[
                    self.query_adata.obs.query(
                        create_query(self.list_clusters, self.use_label)
                    ).index
                ].copy()
            else:
                self.query_adata = self.query_adata.copy()
        else:
            self.query_adata = self.query_adata.copy()

    def _generate_frame(self):
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots(figsize=self.figsize)
        return [fig, ax]

    def _add_image(self, main_ax: Axes):
        image = self.query_adata.uns["spatial"][self.library_id]["images"][self.img_key]
        main_ax.imshow(
            image,
            alpha=self.image_alpha,
            zorder=-1,
        )

    def _plot_colorbar(self, plot_ax: Axes, color_bar_label: str = ""):

        cb = plt.colorbar(
            plot_ax, aspect=10, shrink=0.5, cmap=self.cmap, label=color_bar_label
        )
        cb.outline.set_visible(False)

    def _remove_axis(self, main_ax: Axes):
        main_ax.axis("off")

    def _crop_image(self, main_ax: _AxesSubplot, margin: float):

        main_ax.set_xlim(self.imagecol.min() - margin, self.imagecol.max() + margin)

        main_ax.set_ylim(self.imagerow.min() - margin, self.imagerow.max() + margin)

        main_ax.set_ylim(main_ax.get_ylim()[::-1])

    def _zoom_image(self, main_ax: _AxesSubplot, zoom_coord: Optional[float]):

        main_ax.set_xlim(zoom_coord[0], zoom_coord[1])
        main_ax.set_ylim(zoom_coord[2], zoom_coord[3])

    def _add_color_bar(self, plot, color_bar_label: str = ""):
        cb = plt.colorbar(
            plot,
            aspect=10,
            shrink=0.5,
            cmap=self.cmap,
            label=color_bar_label,
        )
        cb.outline.set_visible(False)

    def _add_title(self):
        plt.title(self.title)

    def _get_query_clusters_index(self):
        index_query = []
        full_labels = self.adata[0].obs[self.use_label].cat.categories

        for query in self.list_clusters:
            index_query.append(np.where(np.array(full_labels) == query)[0][0])

        return index_query

    def _save_output(self):

        self.fig.savefig(
            fname=self.fname, bbox_inches="tight", pad_inches=0, dpi=self.dpi
        )


################################################################
#                                                              #
#                      Gene plot class                         #
#                                                              #
################################################################

import warnings


class GenePlot(SpatialBasePlot):
    def __init__(
        self,
        adata: AnnData,
        # plotting param
        title: Optional["str"] = None,
        figsize: Optional[Tuple[float, float]] = None,
        cmap: Optional[str] = "Spectral_r",
        use_label: Optional[str] = None,
        list_clusters: Optional[list] = None,
        ax: Optional[matplotlib.axes._subplots.Axes] = None,
        fig: Optional[matplotlib.figure.Figure] = None,
        show_plot: Optional[bool] = True,
        show_axis: Optional[bool] = False,
        show_image: Optional[bool] = True,
        show_color_bar: Optional[bool] = True,
        color_bar_label: Optional[str] = "",
        crop: Optional[bool] = True,
        zoom_coord: Optional[float] = None,
        margin: Optional[bool] = 100,
        size: Optional[float] = 7,
        image_alpha: Optional[float] = 1.0,
        cell_alpha: Optional[float] = 1.0,
        use_raw: Optional[bool] = False,
        fname: Optional[str] = None,
        dpi: Optional[int] = 120,
        # gene plot param
        gene_symbols: Union[str, list] = None,
        threshold: Optional[float] = None,
        method: str = "CumSum",
        contour: bool = False,
        step_size: Optional[int] = None,
        vmin: float = None,
        vmax: float = None,
        **kwargs,
    ):
        super().__init__(
            adata=adata,
            title=title,
            figsize=figsize,
            cmap=cmap,
            use_label=use_label,
            list_clusters=list_clusters,
            ax=ax,
            fig=fig,
            show_plot=show_plot,
            show_axis=show_axis,
            show_image=show_image,
            show_color_bar=show_color_bar,
            zoom_coord=zoom_coord,
            crop=crop,
            margin=margin,
            size=size,
            image_alpha=image_alpha,
            cell_alpha=cell_alpha,
            use_raw=use_raw,
            fname=fname,
            dpi=dpi,
        )

        method_available = ["CumSum", "NaiveMean"]
        assert method in method_available, "Please choose available method in: " + str(
            method_available
        )
        self.method = method

        self.step_size = step_size

        if self.title == None:
            if type(gene_symbols) == str:

                self.title = str(gene_symbols)
                gene_symbols = [gene_symbols]
            else:
                self.title = ", ".join(gene_symbols)

        self._add_title()

        self.gene_symbols = gene_symbols

        gene_values = self._get_gene_expression()

        self.available_ids = self._add_threshold(gene_values, threshold)

        self.vmin, self.vmax = vmin, vmax

        if contour:
            plot = self._plot_contour(gene_values[self.available_ids])
        else:
            plot = self._plot_genes(gene_values[self.available_ids])

        if show_color_bar:
            self._add_color_bar(plot, color_bar_label=color_bar_label)

        if fname != None:
            self._save_output()

    def _get_gene_expression(self):

        # Gene plot option
        if len(self.gene_symbols) == 0:
            raise ValueError("Genes should be provided, please input genes")

        elif len(self.gene_symbols) == 1:

            if self.gene_symbols[0] not in self.query_adata.var_names:
                raise ValueError(
                    self.gene_symbols[0]
                    + " is not exist in the data, please try another gene"
                )

            colors = self.query_adata[:, self.gene_symbols].to_df().iloc[:, -1]

            return colors
        else:

            for gene in self.gene_symbols:
                if gene not in self.query_adata.var.index:
                    self.gene_symbols.remove(gene)
                    warnings.warn(
                        "We removed " + gene + " because they not exist in the data"
                    )
                if len(self.gene_symbols) == 0:
                    raise ValueError("All provided genes are not exist in the data")

            count_gene = self.query_adata[:, self.gene_symbols].to_df()

            if self.method is None:
                raise ValueError(
                    "Please provide method to combine genes by NaiveMean/CumSum"
                )

            if self.method == "NaiveMean":
                present_genes = (count_gene > 0).sum(axis=1) / len(self.gene_symbols)

                count_gene = (count_gene.mean(axis=1)) * present_genes

            elif self.method == "CumSum":
                count_gene = count_gene.cumsum(axis=1).iloc[:, -1]

            colors = count_gene

            return colors

    def _plot_genes(self, gene_values: pd.Series):

        if type(self.vmin) == type(None) and type(self.vmax) == type(None):
            vmin = min(gene_values)
            vmax = max(gene_values)
        else:
            vmin, vmax = self.vmin, self.vmax

        # Plot scatter plot based on pixel of spots
        imgcol_new = self.query_adata.obsm["spatial"][:, 0] * self.scale_factor
        imgrow_new = self.query_adata.obsm["spatial"][:, 1] * self.scale_factor
        plot = self.ax.scatter(
            imgcol_new,
            imgrow_new,
            edgecolor="none",
            alpha=self.cell_alpha,
            s=self.size,
            marker="o",
            vmin=vmin,
            vmax=vmax,
            cmap=plt.get_cmap(self.cmap) if type(self.cmap) == str else self.cmap,
            c=gene_values,
        )
        return plot

    def _plot_contour(self, gene_values: pd.Series):

        imgcol_new = self.query_adata.obsm["spatial"][:, 0] * self.scale_factor
        imgrow_new = self.query_adata.obsm["spatial"][:, 1] * self.scale_factor
        # Extracting x,y and values (z)
        z = gene_values
        y = imgrow_new
        x = imgcol_new

        # Interpolating values to get better coverage
        xi = np.linspace(x.min(), x.max(), 100)
        yi = np.linspace(y.min(), y.max(), 100)
        zi = griddata((x, y), z, (xi[None, :], yi[:, None]), method="linear")

        if self.step_size == None:
            self.step_size = int(np.max(z) / 50)
            if self.step_size < 1:
                self.step_size = 1
        # Creating contour plot with a step size of 1

        cs = plt.contourf(
            xi,
            yi,
            zi,
            range(0, int(np.nanmax(zi)) + self.step_size, self.step_size),
            cmap=plt.get_cmap(self.cmap) if type(self.cmap) == str else self.cmap,
            alpha=self.cell_alpha,
        )
        return cs

    def _add_threshold(self, gene_values, threshold):
        if threshold == None:
            return np.repeat(True, len(gene_values))
        else:
            return gene_values > threshold


################################################################
#                                                              #
#                      Feature plot class                      #
#                                                              #
################################################################


class FeaturePlot(SpatialBasePlot):
    def __init__(
        self,
        adata: AnnData,
        # plotting param
        title: Optional["str"] = None,
        figsize: Optional[Tuple[float, float]] = None,
        cmap: Optional[str] = "Spectral_r",
        use_label: Optional[str] = None,
        list_clusters: Optional[list] = None,
        ax: Optional[matplotlib.axes._subplots.Axes] = None,
        fig: Optional[matplotlib.figure.Figure] = None,
        show_plot: Optional[bool] = True,
        show_axis: Optional[bool] = False,
        show_image: Optional[bool] = True,
        show_color_bar: Optional[bool] = True,
        color_bar_label: Optional[str] = "",
        crop: Optional[bool] = True,
        zoom_coord: Optional[float] = None,
        margin: Optional[bool] = 100,
        size: Optional[float] = 7,
        image_alpha: Optional[float] = 1.0,
        cell_alpha: Optional[float] = 1.0,
        use_raw: Optional[bool] = False,
        fname: Optional[str] = None,
        dpi: Optional[int] = 120,
        # gene plot param
        feature: str = None,
        threshold: Optional[float] = None,
        contour: bool = False,
        step_size: Optional[int] = None,
        vmin: float = None,
        vmax: float = None,
        **kwargs,
    ):
        super().__init__(
            adata=adata,
            title=title,
            figsize=figsize,
            cmap=cmap,
            use_label=use_label,
            list_clusters=list_clusters,
            ax=ax,
            fig=fig,
            show_plot=show_plot,
            show_axis=show_axis,
            show_image=show_image,
            show_color_bar=show_color_bar,
            zoom_coord=zoom_coord,
            crop=crop,
            margin=margin,
            size=size,
            image_alpha=image_alpha,
            cell_alpha=cell_alpha,
            use_raw=use_raw,
            fname=fname,
            dpi=dpi,
        )

        self.step_size = step_size

        self.title = feature
        self._add_title()

        self.feature = feature

        feature_values = self._get_feature_values()

        self.available_ids = self._add_threshold(feature_values, threshold)

        self.vmin, self.vmax = vmin, vmax

        if contour:
            plot = self._plot_contour(feature_values[self.available_ids])
        else:
            plot = self._plot_feature(feature_values[self.available_ids])

        if show_color_bar:
            self._add_color_bar(plot, color_bar_label=color_bar_label)

        if fname != None:
            self._save_output()

    def _get_feature_values(self):

        if self.feature not in self.query_adata.obs:
            raise ValueError(
                self.feature + " is not in data.obs, please try another feature"
            )
        elif not isinstance(
            self.query_adata.obs[self.feature].values[0], numbers.Number
        ):
            raise ValueError(
                self.feature
                + " in data.obs is not continuous, please try another feature"
            )

        colors = self.query_adata.obs[self.feature]

        return colors

    def _plot_feature(self, feature_values: pd.Series):

        if type(self.vmin) == type(None) and type(self.vmax) == type(None):
            vmin = min(feature_values)
            vmax = max(feature_values)
        else:
            vmin, vmax = self.vmin, self.vmax

        # Plot scatter plot based on pixel of spots
        imgcol_new = self.query_adata.obsm["spatial"][:, 0] * self.scale_factor
        imgrow_new = self.query_adata.obsm["spatial"][:, 1] * self.scale_factor
        plot = self.ax.scatter(
            imgcol_new,
            imgrow_new,
            edgecolor="none",
            alpha=self.cell_alpha,
            s=self.size,
            marker="o",
            vmin=vmin,
            vmax=vmax,
            cmap=plt.get_cmap(self.cmap) if type(self.cmap) == str else self.cmap,
            c=feature_values,
        )
        return plot

    def _plot_contour(self, feature_values: pd.Series):

        imgcol_new = self.query_adata.obsm["spatial"][:, 0] * self.scale_factor
        imgrow_new = self.query_adata.obsm["spatial"][:, 1] * self.scale_factor
        # Extracting x,y and values (z)
        z = feature_values
        y = imgrow_new
        x = imgcol_new

        # Interpolating values to get better coverage
        xi = np.linspace(x.min(), x.max(), 100)
        yi = np.linspace(y.min(), y.max(), 100)
        zi = griddata((x, y), z, (xi[None, :], yi[:, None]), method="linear")

        if self.step_size == None:
            self.step_size = int(np.max(z) / 50)
            if self.step_size < 1:
                self.step_size = 1
        # Creating contour plot with a step size of 1

        cs = plt.contourf(
            xi,
            yi,
            zi,
            range(0, int(np.nanmax(zi)) + self.step_size, self.step_size),
            cmap=plt.get_cmap(self.cmap) if type(self.cmap) == str else self.cmap,
            alpha=self.cell_alpha,
        )
        return cs

    def _add_threshold(self, feature_values, threshold):
        if threshold == None:
            return np.repeat(True, len(feature_values))
        else:
            return feature_values > threshold


################################################################
#                                                              #
#                      Cluster plot class                      #
#                                                              #
################################################################


class ClusterPlot(SpatialBasePlot):
    def __init__(
        self,
        adata: AnnData,
        # plotting param
        title: Optional["str"] = None,
        figsize: Optional[Tuple[float, float]] = None,
        cmap: Optional[str] = "default",
        use_label: Optional[str] = None,
        list_clusters: Optional[list] = None,
        ax: Optional[matplotlib.axes._subplots.Axes] = None,
        fig: Optional[matplotlib.figure.Figure] = None,
        show_plot: Optional[bool] = True,
        show_axis: Optional[bool] = False,
        show_image: Optional[bool] = True,
        show_color_bar: Optional[bool] = True,
        crop: Optional[bool] = True,
        zoom_coord: Optional[float] = None,
        margin: Optional[bool] = 100,
        size: Optional[float] = 5,
        image_alpha: Optional[float] = 1.0,
        cell_alpha: Optional[float] = 1.0,
        fname: Optional[str] = None,
        dpi: Optional[int] = 120,
        # cluster plot param
        show_subcluster: Optional[bool] = False,
        show_cluster_labels: Optional[bool] = False,
        show_trajectories: Optional[bool] = False,
        reverse: Optional[bool] = False,
        show_node: Optional[bool] = False,
        threshold_spots: Optional[int] = 5,
        text_box_size: Optional[float] = 5,
        color_bar_size: Optional[float] = 10,
        bbox_to_anchor: Optional[Tuple[float, float]] = (1, 1),
        # trajectory
        trajectory_node_size: Optional[int] = 10,
        trajectory_alpha: Optional[float] = 1.0,
        trajectory_width: Optional[float] = 2.5,
        trajectory_edge_color: Optional[str] = "#f4efd3",
        trajectory_arrowsize: Optional[int] = 17,
    ):
        super().__init__(
            adata=adata,
            title=title,
            figsize=figsize,
            cmap=cmap,
            use_label=use_label,
            list_clusters=list_clusters,
            ax=ax,
            fig=fig,
            show_plot=show_plot,
            show_axis=show_axis,
            show_image=show_image,
            show_color_bar=show_color_bar,
            zoom_coord=zoom_coord,
            crop=crop,
            margin=margin,
            size=size,
            image_alpha=image_alpha,
            cell_alpha=cell_alpha,
            fname=fname,
            dpi=dpi,
        )

        self.cmap_ = self._get_cmap(self.cmap)

        self._add_cluster_colors()

        self._plot_clusters()

        self.threshold_spots = threshold_spots
        self.text_box_size = text_box_size
        self.color_bar_size = color_bar_size
        self.reverse = reverse
        self.show_node = show_node

        if show_color_bar:
            self._add_cluster_bar(bbox_to_anchor)

        if show_cluster_labels:
            self._add_cluster_labels()

        if show_subcluster:
            self._add_sub_clusters()

        if show_trajectories:

            self.trajectory_node_size = trajectory_node_size
            self.trajectory_alpha = trajectory_alpha
            self.trajectory_width = trajectory_width
            self.trajectory_edge_color = trajectory_edge_color
            self.trajectory_arrowsize = trajectory_arrowsize

            self._add_trajectories()

        if fname != None:
            self._save_output()

    def _add_cluster_colors(self):
        if self.use_label + "_colors" not in self.adata[0].uns:
            # self.adata[0].uns[self.use_label + "_set"] = []
            self.adata[0].uns[self.use_label + "_colors"] = []

            for i, cluster in enumerate(self.adata[0].obs.groupby(self.use_label)):
                self.adata[0].uns[self.use_label + "_colors"].append(
                    matplotlib.colors.to_hex(self.cmap_(i / (self.cmap_n - 1)))
                )
                # self.adata[0].uns[self.use_label + "_set"].append( cluster[0] )

    def _plot_clusters(self):
        # Plot scatter plot based on pixel of spots

        # for i, cluster in enumerate(self.query_adata.obs[self.use_label].cat.categories):
        for i, cluster in enumerate(self.query_adata.obs.groupby(self.use_label)):

            # Plot scatter plot based on pixel of spots
            subset_spatial = self.query_adata.obsm["spatial"][
                check_sublist(list(self.query_adata.obs.index), list(cluster[1].index))
            ]

            if self.use_label + "_colors" in self.adata[0].uns:
                # label_set = self.adata[0].uns[self.use_label+'_set']
                label_set = (
                    self.adata[0].obs[self.use_label].cat.categories.values.astype(str)
                )
                col_index = np.where(label_set == cluster[0])[0][0]
                color = self.adata[0].uns[self.use_label + "_colors"][col_index]
            else:
                color = self.cmap_(self.query_indexes[i] / (self.cmap_n - 1))

            imgcol_new = subset_spatial[:, 0] * self.scale_factor
            imgrow_new = subset_spatial[:, 1] * self.scale_factor
            _ = self.ax.scatter(
                imgcol_new,
                imgrow_new,
                c=[color],
                label=cluster[0],
                edgecolor="none",
                alpha=self.cell_alpha,
                s=self.size,
                marker="o",
            )

    def _get_cmap(self, cmap):
        cmap_, cmap_n = get_cmap(cmap)
        self.cmap_n = cmap_n
        return cmap_

    def _add_cluster_bar(self, bbox_to_anchor):
        lgnd = self.ax.legend(
            bbox_to_anchor=bbox_to_anchor,
            labelspacing=0.05,
            fontsize=self.color_bar_size,
            handleheight=1.0,
            edgecolor="white",
        )
        for handle in lgnd.legendHandles:
            handle.set_sizes([20.0])

    def _add_cluster_labels(self):

        for i, label in enumerate(self.list_clusters):

            label_index = list(
                self.query_adata.obs[
                    self.query_adata.obs[self.use_label] == str(label)
                ].index
            )
            subset_spatial = self.query_adata.obsm["spatial"][
                check_sublist(list(self.query_adata.obs.index), label_index)
            ]

            imgcol_new = subset_spatial[:, 0] * self.scale_factor
            imgrow_new = subset_spatial[:, 1] * self.scale_factor

            centroids = [centroidpython(imgcol_new, imgrow_new)]

            if centroids[0][0] < 1500:
                x = -100
                y = 50
            else:
                x = 100
                y = -50

            colors = self.adata[0].uns[self.use_label + "_colors"]
            index = self.query_indexes[i]
            self.ax.text(
                centroids[0][0] + x,
                centroids[0][1] + y,
                label,
                color="black",
                fontsize=self.text_box_size,
                zorder=3,
                bbox=dict(
                    facecolor=colors[index],
                    boxstyle="round",
                    alpha=1.0,
                ),
            )

    def _add_sub_clusters(self):

        if "sub_cluster_labels" not in self.query_adata.obs.columns:
            raise ValueError("Please run stlearn.spatial.cluster.localization")

        for i, label in enumerate(self.list_clusters):
            label_index = list(
                self.query_adata.obs[
                    self.query_adata.obs[self.use_label] == str(label)
                ].index
            )
            subset_spatial = self.query_adata.obsm["spatial"][
                check_sublist(list(self.query_adata.obs.index), label_index)
            ]

            imgcol_new = subset_spatial[:, 0] * self.scale_factor
            imgrow_new = subset_spatial[:, 1] * self.scale_factor

            if (
                len(
                    self.query_adata.obs[
                        self.query_adata.obs[self.use_label] == str(label)
                    ]["sub_cluster_labels"].unique()
                )
                < 2
            ):
                centroids = [centroidpython(imgcol_new, imgrow_new)]
                classes = np.array(
                    self.query_adata.obs[
                        self.query_adata.obs[self.use_label] == str(label)
                    ]["sub_cluster_labels"].unique()
                )

            else:
                from sklearn.neighbors import NearestCentroid

                clf = NearestCentroid()
                clf.fit(
                    np.column_stack((imgcol_new, imgrow_new)),
                    self.query_adata.obs[
                        self.query_adata.obs[self.use_label] == str(label)
                    ]["sub_cluster_labels"],
                )

                centroids = clf.centroids_
                classes = clf.classes_

            for j, label in enumerate(classes):
                if (
                    len(
                        self.query_adata.obs[
                            self.query_adata.obs["sub_cluster_labels"] == label
                        ]
                    )
                    > self.threshold_spots
                ):
                    if centroids[j][0] < 1500:
                        x = -100
                        y = 50
                    else:
                        x = 100
                        y = -50
                    colors = self.adata[0].uns[self.use_label + "_colors"]
                    index = self.query_indexes[i]

                    self.ax.text(
                        centroids[j][0] + x,
                        centroids[j][1] + y,
                        label,
                        color="black",
                        fontsize=5,
                        zorder=3,
                        bbox=dict(
                            facecolor=colors[index],
                            boxstyle="round",
                            alpha=1.0,
                        ),
                        clip_on=True,
                    )

    def _add_trajectories(self):
        used_colors = self.adata[0].uns[self.use_label + "_colors"]
        cmaps = matplotlib.colors.LinearSegmentedColormap.from_list("", used_colors)

        cmap = plt.get_cmap(cmaps)

        if "PTS_graph" not in self.adata[0].uns:
            raise ValueError("Please run stlearn.spatial.trajectory.pseudotimespace!")

        tmp = _read_graph(self.adata[0], "PTS_graph")

        G = tmp.copy()

        remove = [edge for edge in G.edges if 9999 in edge]
        G.remove_edges_from(remove)
        G.remove_node(9999)
        centroid_dict = self.adata[0].uns["centroid_dict"]
        centroid_dict = {int(key): centroid_dict[key] for key in centroid_dict}
        if self.reverse:
            nx.draw_networkx_edges(
                G,
                pos=centroid_dict,
                node_size=self.trajectory_node_size,
                alpha=self.trajectory_alpha,
                width=self.trajectory_width,
                edge_color=self.trajectory_edge_color,
                arrowsize=self.trajectory_arrowsize,
                arrowstyle="<|-",
                connectionstyle="arc3,rad=0.2",
            )
        else:
            nx.draw_networkx_edges(
                G,
                pos=centroid_dict,
                node_size=self.trajectory_node_size,
                alpha=self.trajectory_alpha,
                width=self.trajectory_width,
                edge_color=self.trajectory_edge_color,
                arrowsize=self.trajectory_arrowsize,
                arrowstyle="-|>",
                connectionstyle="arc3,rad=0.2",
            )

        if self.show_node:
            for x, y in centroid_dict.items():

                if x in get_node(self.list_clusters, self.adata[0].uns["split_node"]):
                    self.ax.text(
                        y[0],
                        y[1],
                        get_cluster(str(x), self.adata[0].uns["split_node"]),
                        color="black",
                        fontsize=8,
                        zorder=100,
                        bbox=dict(
                            facecolor=cmap(
                                int(
                                    get_cluster(str(x), self.adata[0].uns["split_node"])
                                )
                                / (len(used_colors) - 1)
                            ),
                            boxstyle="circle",
                            alpha=1,
                        ),
                    )


################################################################
#                                                              #
#                      SubCluster plot class                   #
#                                                              #
################################################################


class SubClusterPlot(SpatialBasePlot):
    def __init__(
        self,
        adata: AnnData,
        # plotting param
        title: Optional["str"] = None,
        figsize: Optional[Tuple[float, float]] = None,
        cmap: Optional[str] = "jet",
        use_label: Optional[str] = None,
        list_clusters: Optional[list] = None,
        ax: Optional[matplotlib.axes._subplots.Axes] = None,
        fig: Optional[matplotlib.figure.Figure] = None,
        show_plot: Optional[bool] = True,
        show_axis: Optional[bool] = False,
        show_image: Optional[bool] = True,
        show_color_bar: Optional[bool] = True,
        crop: Optional[bool] = True,
        zoom_coord: Optional[float] = None,
        margin: Optional[bool] = 100,
        size: Optional[float] = 5,
        image_alpha: Optional[float] = 1.0,
        cell_alpha: Optional[float] = 1.0,
        fname: Optional[str] = None,
        dpi: Optional[int] = 120,
        # subcluster plot param
        cluster: Optional[int] = 0,
        threshold_spots: Optional[int] = 5,
        text_box_size: Optional[float] = 5,
        bbox_to_anchor: Optional[Tuple[float, float]] = (1, 1),
        **kwargs,
    ):
        super().__init__(
            adata=adata,
            title=title,
            figsize=figsize,
            cmap=cmap,
            use_label=use_label,
            list_clusters=list_clusters,
            ax=ax,
            fig=fig,
            show_plot=show_plot,
            show_axis=show_axis,
            show_image=show_image,
            show_color_bar=show_color_bar,
            zoom_coord=zoom_coord,
            crop=crop,
            margin=margin,
            size=size,
            image_alpha=image_alpha,
            cell_alpha=cell_alpha,
            fname=fname,
            dpi=dpi,
        )

        self.text_box_size = text_box_size
        self.cluster = cluster

        subset = self._plot_subclusters(threshold_spots)

        self._add_subclusters_label(subset)

        if fname != None:
            self._save_output()

    def _plot_subclusters(self, threshold_spots):
        subset = (
            self.adata[0]
            .obs[self.adata[0].obs[self.use_label] == str(self.cluster)]
            .copy()
        )

        meaningful_sub = []
        for i in subset["sub_cluster_labels"].unique():
            if len(subset[subset["sub_cluster_labels"] == str(i)]) > threshold_spots:
                meaningful_sub.append(i)

        subset = subset[subset["sub_cluster_labels"].isin(meaningful_sub)]

        colors = subset["sub_cluster_labels"]
        sub_anndata = self.adata[0][subset.index, :].copy()
        self.imgcol_new = sub_anndata.obsm["spatial"][:, 0] * self.scale_factor
        self.imgrow_new = sub_anndata.obsm["spatial"][:, 1] * self.scale_factor

        keys = list(np.sort(colors.unique()))
        self.vals = np.arange(len(keys))
        self.mapping = dict(zip(keys, self.vals))

        colors = colors.replace(self.mapping)

        plot = self.ax.scatter(
            self.imgcol_new,
            self.imgrow_new,
            edgecolor="none",
            s=self.size,
            marker="o",
            cmap=plt.get_cmap(self.cmap) if type(self.cmap) == str else self.cmap,
            c=colors,
            alpha=self.cell_alpha,
        )

        return subset

    def _add_subclusters_label(self, subset):
        if len(subset["sub_cluster_labels"].unique()) < 2:
            print("lower than 2")
            centroids = [centroidpython(subset[["imagecol", "imagerow"]].values)]
            classes = np.array([subset["sub_cluster_labels"][0]])

        else:
            from sklearn.neighbors import NearestCentroid

            clf = NearestCentroid()
            clf.fit(
                np.column_stack((self.imgcol_new, self.imgrow_new)),
                subset["sub_cluster_labels"],
            )

            centroids = clf.centroids_
            classes = clf.classes_

        norm = matplotlib.colors.Normalize(vmin=min(self.vals), vmax=max(self.vals))

        m = matplotlib.cm.ScalarMappable(norm=norm, cmap=self.cmap)

        for i, label in enumerate(classes):
            if centroids[i][0] < 1000:
                x = -100
                y = 100
            else:
                x = 100
                y = -100

            self.ax.text(
                centroids[i][0] + x,
                centroids[i][1] + y,
                label,
                color="white",
                fontsize=self.text_box_size,
                zorder=3,
                bbox=dict(
                    facecolor=matplotlib.colors.to_hex(m.to_rgba(self.mapping[label])),
                    boxstyle="round",
                    alpha=0.5,
                ),
            )


################################################################
#                                                              #
#                      Cci Plot class                          #
#                                                              #
################################################################


class CciPlot(GenePlot):
    def __init__(
        self,
        adata: AnnData,
        # plotting param
        title: Optional["str"] = None,
        figsize: Optional[Tuple[float, float]] = None,
        cmap: Optional[str] = "Spectral_r",
        use_label: Optional[str] = None,
        list_clusters: Optional[list] = None,
        ax: Optional[matplotlib.axes._subplots.Axes] = None,
        fig: Optional[matplotlib.figure.Figure] = None,
        show_plot: Optional[bool] = True,
        show_axis: Optional[bool] = False,
        show_image: Optional[bool] = True,
        show_color_bar: Optional[bool] = True,
        crop: Optional[bool] = True,
        zoom_coord: Optional[float] = None,
        margin: Optional[bool] = 100,
        size: Optional[float] = 7,
        image_alpha: Optional[float] = 1.0,
        cell_alpha: Optional[float] = 1.0,
        use_raw: Optional[bool] = False,
        fname: Optional[str] = None,
        dpi: Optional[int] = 120,
        # cci_rank param
        use_het: Optional[str] = "het",
        contour: bool = False,
        step_size: Optional[int] = None,
        vmin: float = None,
        vmax: float = None,
        **kwargs,
    ):
        super().__init__(
            adata=adata,
            figsize=figsize,
            cmap=cmap,
            use_label=use_label,
            list_clusters=list_clusters,
            ax=ax,
            fig=fig,
            show_plot=show_plot,
            show_axis=show_axis,
            show_image=show_image,
            show_color_bar=show_color_bar,
            zoom_coord=zoom_coord,
            crop=crop,
            margin=margin,
            size=size,
            image_alpha=image_alpha,
            cell_alpha=cell_alpha,
            use_raw=use_raw,
            fname=fname,
            dpi=dpi,
            gene_symbols=use_het,
            contour=contour,
            step_size=step_size,
            vmin=vmin,
            vmax=vmax,
        )

        self.title = title

        self._add_title()

    def _get_gene_expression(self):
        return self.query_adata.obsm[self.gene_symbols[0]]


class LrResultPlot(GenePlot):
    def __init__(
        self,
        adata: AnnData,
        use_lr: Optional["str"] = None,
        use_result: Optional["str"] = "lr_sig_scores",
        # plotting param
        title: Optional["str"] = None,
        figsize: Optional[Tuple[float, float]] = None,
        cmap: Optional[str] = "Spectral_r",
        list_clusters: Optional[list] = None,
        ax: Optional[matplotlib.axes._subplots.Axes] = None,
        fig: Optional[matplotlib.figure.Figure] = None,
        show_plot: Optional[bool] = True,
        show_axis: Optional[bool] = False,
        show_image: Optional[bool] = True,
        show_color_bar: Optional[bool] = True,
        crop: Optional[bool] = True,
        zoom_coord: Optional[float] = None,
        margin: Optional[bool] = 100,
        size: Optional[float] = 7,
        image_alpha: Optional[float] = 1.0,
        cell_alpha: Optional[float] = 1.0,
        use_raw: Optional[bool] = False,
        fname: Optional[str] = None,
        dpi: Optional[int] = 120,
        # cci_rank param
        contour: bool = False,
        step_size: Optional[int] = None,
        vmin: float = None,
        vmax: float = None,
        **kwargs,
    ):
        # Making sure cci_rank has been run first #
        if "lr_summary" not in adata.uns:
            raise Exception(
                f"To visualise LR interaction results, must run" f"st.pl.cci.run first."
            )

        # By default, using the LR with most significant spots #
        if type(use_lr) == type(None):
            use_lr = adata.uns["lr_summary"].index.values[0]
        elif use_lr not in adata.uns["lr_summary"].index:
            raise Exception(
                f"use_lr must be one of:\n" f'{adata.uns["lr_summary"].index}'
            )
        else:
            use_lr = str(use_lr)

        # Checking is a valid result #
        res_info = ["lr_scores", "p_vals", "p_adjs", "-log10(p_adjs)", "lr_sig_scores"]
        if use_result not in res_info:
            raise Exception(f"use_result must be one of:\n{res_info}")
        else:
            self.use_result = use_result

        super().__init__(
            adata=adata,
            title=title,
            figsize=figsize,
            cmap=cmap,
            list_clusters=list_clusters,
            ax=ax,
            fig=fig,
            show_plot=show_plot,
            show_axis=show_axis,
            show_image=show_image,
            show_color_bar=show_color_bar,
            zoom_coord=zoom_coord,
            crop=crop,
            margin=margin,
            size=size,
            image_alpha=image_alpha,
            cell_alpha=cell_alpha,
            use_raw=use_raw,
            fname=fname,
            dpi=dpi,
            gene_symbols=use_lr,
            contour=contour,
            step_size=step_size,
            vmin=vmin,
            vmax=vmax,
        )

    def _get_gene_expression(self):
        use_lr = self.gene_symbols[0]
        index = np.where(self.query_adata.uns["lr_summary"].index.values == use_lr)[0][
            0
        ]
        return self.query_adata.obsm[self.use_result][:, index]



import sys, os, random, scipy
import numpy as np
import pandas as pd
from numba.typed import List
import statsmodels.api as sm
from statsmodels.stats.multitest import multipletests

from tqdm import tqdm
from sklearn.cluster import AgglomerativeClustering

from anndata import AnnData
from .base import lr, calc_neighbours, get_spot_lrs, get_lrs_scores, get_scores
from .merge import merge
from .perm_utils import get_lr_features, get_lr_bg

# Newest method #
def perform_spot_testing(
    adata: AnnData,
    lr_scores: np.ndarray,
    lrs: np.array,
    n_pairs: int,
    neighbours: List,
    het_vals: np.array,
    min_expr: float,
    adj_method: str = "fdr_bh",
    pval_adj_cutoff: float = 0.05,
    verbose: bool = True,
    save_bg=False,
    neg_binom=False,
    quantiles=(0.5, 0.75, 0.85, 0.9, 0.95, 0.97, 0.98, 0.99, 0.995, 0.9975, 0.999, 1),
):
    """Calls significant spots by creating random gene pairs with similar
    expression to given LR pair; only generate background for spots
    which have score for given LR.
    """
    quantiles = np.array(quantiles)

    lr_genes = np.unique([lr_.split("_") for lr_ in lrs])
    genes = np.array([gene for gene in adata.var_names if gene not in lr_genes])
    candidate_expr = adata[:, genes].to_df().values

    n_genes = round(np.sqrt(n_pairs) * 2)
    if len(genes) < n_genes:
        print(
            "Exiting since need atleast "
            f"{n_genes} genes to generate {n_pairs} pairs."
        )
        return

    if n_pairs < 100:
        print(
            "Exiting since n_pairs<100, need much larger number of pairs to "
            "get accurate backgrounds (e.g. 1000)."
        )
        return

    ####### Quantiles to select similar gene to LRs to gen. rand-pairs #######
    lr_expr = adata[:, lr_genes].to_df()
    lr_feats = get_lr_features(adata, lr_expr, lrs, quantiles)
    l_quants = lr_feats.loc[
        lrs, [col for col in lr_feats.columns if "L_" in col]
    ].values
    r_quants = lr_feats.loc[
        lrs, [col for col in lr_feats.columns if "R_" in col]
    ].values
    candidate_quants = np.apply_along_axis(
        np.quantile, 0, candidate_expr, q=quantiles, interpolation="nearest"
    )
    # Ensuring consistent typing to prevent numba errors #
    l_quants = l_quants.astype("<f4")
    r_quants = r_quants.astype("<f4")
    candidate_quants = candidate_quants.astype("<f4")

    ######## Background per LR, but only for spots where LR has a score ########
    # Determine the indices of the spots where each LR has a score #
    cols = ["n_spots", "n_spots_sig", "n_spots_sig_pval"]
    lr_summary = np.zeros((lr_scores.shape[1], 3), np.int)
    pvals = np.ones(lr_scores.shape, dtype=np.float64)
    pvals_adj = np.ones(lr_scores.shape, dtype=np.float64)
    log10pvals_adj = np.zeros(lr_scores.shape, dtype=np.float64)
    lr_sig_scores = lr_scores.copy()

    # If we are saving the backgrounds #
    if save_bg:
        adata.uns["lrs_to_bg"] = {}
        adata.uns["lr_spot_indices"] = {}

    with tqdm(
        total=lr_scores.shape[1],
        desc="Generating backgrounds & testing each LR pair...",
        bar_format="{l_bar}{bar} [ time left: {remaining} ]",
        disable=verbose == False,
    ) as pbar:

        gene_bg_genes = {}  # Keep track of genes which can be used to gen. rand-pairs.
        spot_lr_indices = [
            [] for i in range(lr_scores.shape[0])
        ]  # tracks the lrs tested in a given spot for MHT !!!!
        for lr_j in range(lr_scores.shape[1]):
            lr_ = lrs[lr_j]

            background, spot_indices = get_lr_bg(
                adata,
                neighbours,
                het_vals,
                min_expr,
                lr_,
                lr_scores[:, lr_j],
                l_quants[lr_j, :],
                r_quants[lr_j, :],
                genes,
                candidate_quants,
                gene_bg_genes,
                n_genes,
                n_pairs,
            )

            if save_bg:
                adata.uns["lrs_to_bg"][lr_] = background
                adata.uns["lr_spot_indices"][lr_] = spot_indices

            if not neg_binom:  # Calculate empirical p-values per-spot
                for spot_i, spot_index in enumerate(spot_indices):
                    n_greater = len(
                        np.where(background[spot_i, :] >= lr_scores[spot_index, lr_j])[
                            0
                        ]
                    )
                    n_greater = n_greater if n_greater != 0 else 1  # pseudocount
                    pvals[spot_index, lr_j] = n_greater / background.shape[1]
                    spot_lr_indices[spot_index].append(lr_j)
            else:  # Fitting NB per LR
                lr_j_scores = lr_scores[spot_indices, lr_j]
                bg_ = background.ravel()
                bg_wScore = np.array(list(lr_j_scores) + list(bg_))

                ##### 1) rounding discretisation
                # First multiple to get minimum value to be one before rounding #
                bg_1 = bg_wScore * (1 / min(bg_wScore[bg_wScore != 0]))
                bg_1 = np.round(bg_1)
                lr_j_scores_1 = bg_1[0 : len(lr_j_scores)]
                bg_1 = bg_1[len(lr_j_scores) : len(bg_1)]

                ###### Getting the pvalue from negative binomial approach
                round_pvals, _, _, _ = get_stats(
                    lr_j_scores_1,
                    bg_1,
                    len(bg_1),
                    neg_binom=True,
                    return_negbinom_params=False,
                )
                pvals[spot_indices, lr_j] = round_pvals
                for spot_index in spot_indices:
                    spot_lr_indices[spot_index].append(lr_j)

            pbar.update(1)

        # MHT correction # filling in other stats #
        for spot_i in range(lr_scores.shape[0]):
            lr_indices = spot_lr_indices[spot_i]
            if len(lr_indices) != 0:
                pvals_adj[spot_i, lr_indices] = multipletests(
                    pvals[spot_i, lr_indices], method=adj_method
                )[1]

            log10pvals_adj[spot_i, :] = -np.log10(pvals_adj[spot_i, :])

            # Recording per lr results for this LR #
            lrs_in_spot = lr_scores[spot_i] > min_expr
            sig_lrs_in_spot = pvals_adj[spot_i, :] < pval_adj_cutoff
            sigpval_lrs_in_spot = pvals[spot_i, :] < pval_adj_cutoff
            lr_summary[lrs_in_spot, 0] += 1
            lr_summary[sig_lrs_in_spot, 1] += 1
            lr_summary[sigpval_lrs_in_spot, 2] += 1

            lr_sig_scores[spot_i, sig_lrs_in_spot == False] = 0

    # Ordering the results according to number of significant spots per LR#
    order = np.argsort(-lr_summary[:, 1])
    lrs_ordered = lrs[order]
    lr_summary = lr_summary[order, :]
    lr_summary = pd.DataFrame(lr_summary, index=lrs_ordered, columns=cols)
    lr_scores = lr_scores[:, order]
    pvals = pvals[:, order]
    pvals_adj = pvals_adj[:, order]
    log10pvals_adj = log10pvals_adj[:, order]
    lr_sig_scores = lr_sig_scores[:, order]

    # Saving the results in AnnData #
    if verbose:
        print("\nStoring results:\n")

    adata.uns["lr_summary"] = lr_summary
    res_info = ["lr_scores", "p_vals", "p_adjs", "-log10(p_adjs)", "lr_sig_scores"]
    mats = [lr_scores, pvals, pvals_adj, log10pvals_adj, lr_sig_scores]
    for i, info_name in enumerate(res_info):
        adata.obsm[info_name] = mats[i]
        if verbose:
            print(f"{info_name} stored in adata.obsm['{info_name}'].")

    if verbose:
        print(
            "\nPer-spot results in adata.obsm have columns in same order as "
            "rows in adata.uns['lr_summary']."
        )
        print("Summary of LR results in adata.uns['lr_summary'].")


# Version 2, no longer in use, see above for newest method #
def perform_perm_testing(
    adata: AnnData,
    lr_scores: np.ndarray,
    n_pairs: int,
    lrs: np.array,
    lr_mid_dist: int,
    verbose: float,
    neighbours: List,
    het_vals: np.array,
    min_expr: float,
    neg_binom: bool,
    adj_method: str,
    pval_adj_cutoff: float,
):
    """Performs the grouped permutation testing when taking the stats approach."""
    if n_pairs != 0:  # Perform permutation testing
        # Grouping spots with similar mean expression point #
        genes = get_valid_genes(adata, n_pairs)
        means_ordered, genes_ordered = get_ordered(adata, genes)
        ims = np.array(
            [
                get_median_index(
                    lr_.split("_")[0],
                    lr_.split("_")[1],
                    means_ordered.values,
                    genes_ordered,
                )
                for lr_ in lrs
            ]
        ).reshape(-1, 1)

        if len(lrs) > 1:  # Multi-LR pair mode, group LRs to generate backgrounds
            clusterer = AgglomerativeClustering(
                n_clusters=None,
                distance_threshold=lr_mid_dist,
                affinity="manhattan",
                linkage="single",
            )
            lr_groups = clusterer.fit_predict(ims)
            lr_group_set = np.unique(lr_groups)
            if verbose:
                print(f"{len(lr_group_set)} lr groups with similar expression levels.")

        else:  # Single LR pair mode, generate background for the LR.
            lr_groups = np.array([0])
            lr_group_set = lr_groups

        res_info = ["lr_scores", "p_val", "p_adj", "-log10(p_adj)", "lr_sig_scores"]
        n_, n_sigs = np.array([0] * len(lrs)), np.array([0] * len(lrs))
        per_lr_results = {}
        with tqdm(
            total=len(lr_group_set),
            desc="Generating background distributions for the LR pair groups..",
            bar_format="{l_bar}{bar} [ time left: {remaining} ]",
        ) as pbar:
            for group in lr_group_set:
                # Determining common mid-point for each group #
                group_bool = lr_groups == group
                group_im = int(np.median(ims[group_bool, 0]))

                # Calculating the background #
                rand_pairs = get_rand_pairs(adata, genes, n_pairs, lrs=lrs, im=group_im)
                background = get_lrs_scores(
                    adata,
                    rand_pairs,
                    neighbours,
                    het_vals,
                    min_expr,
                    filter_pairs=False,
                ).ravel()
                total_bg = len(background)
                background = background[background != 0]  # Filtering for increase speed

                # Getting stats for each lr in group #
                group_lr_indices = np.where(group_bool)[0]
                for lr_i in group_lr_indices:
                    lr_ = lrs[lr_i]
                    lr_results = pd.DataFrame(index=adata.obs_names, columns=res_info)
                    scores = lr_scores[:, lr_i]
                    stats = get_stats(
                        scores,
                        background,
                        total_bg,
                        neg_binom,
                        adj_method,
                        pval_adj_cutoff=pval_adj_cutoff,
                    )
                    full_stats = [scores] + list(stats)
                    for vals, colname in zip(full_stats, res_info):
                        lr_results[colname] = vals

                    n_[lr_i] = len(np.where(scores > 0)[0])
                    n_sigs[lr_i] = len(
                        np.where(lr_results["p_adj"].values < pval_adj_cutoff)[0]
                    )
                    if n_sigs[lr_i] > 0:
                        per_lr_results[lr_] = lr_results
                pbar.update(1)

        print(f"{len(per_lr_results)} LR pairs with significant interactions.")

        lr_summary = pd.DataFrame(index=lrs, columns=["n_spots", "n_spots_sig"])
        lr_summary["n_spots"] = n_
        lr_summary["n_spots_sig"] = n_sigs
        lr_summary = lr_summary.iloc[np.argsort(-n_sigs)]

    else:  # Simply store the scores
        per_lr_results = {}
        lr_summary = pd.DataFrame(index=lrs, columns=["n_spots"])
        for i, lr_ in enumerate(lrs):
            lr_results = pd.DataFrame(index=adata.obs_names, columns=["lr_scores"])
            lr_results["lr_scores"] = lr_scores[:, i]
            per_lr_results[lr_] = lr_results
            lr_summary.loc[lr_, "n_spots"] = len(np.where(lr_scores[:, i] > 0)[0])
        lr_summary = lr_summary.iloc[np.argsort(-lr_summary.values[:, 0]), :]

    adata.uns["lr_summary"] = lr_summary
    adata.uns["per_lr_results"] = per_lr_results
    if verbose:
        print(
            "Summary of significant spots for each lr pair in adata.uns['lr_summary']."
        )
        print(
            "Spot enrichment statistics of LR interactions in adata.uns['per_lr_results']"
        )


# No longer in use #
def permutation(
    adata: AnnData,
    n_pairs: int = 200,
    distance: int = None,
    use_lr: str = "cci_lr",
    use_het: str = None,
    neg_binom: bool = False,
    adj_method: str = "fdr",
    neighbours: list = None,
    run_fast: bool = True,
    bg_pairs: list = None,
    background: np.array = None,
    **kwargs,
) -> AnnData:

    """Permutation test for merged result
    Parameters
    ----------
    adata: AnnData          The data object including the cell types to count
    n_pairs: int            Number of gene pairs to run permutation test (default: 1000)
    distance: int           Distance between spots (default: 30)
    use_lr: str             LR cluster used for permutation test (default: 'lr_neighbours_louvain_max')
    use_het: str            cell type diversity counts used for permutation test (default 'het')
    neg_binom: bool         Whether to fit neg binomial paramaters to bg distribution for p-val est.
    adj_method: str         Method used by statsmodels.stats.multitest.multipletests for MHT correction.
    neighbours: list        List of the neighbours for each spot, if None then computed. Useful for speeding up function.
    **kwargs:               Extra arguments parsed to lr.
    Returns
    -------
    adata: AnnData          Data Frame of p-values from permutation test for each window stored in adata.uns['merged_pvalues']
                            Final significant merged scores stored in adata.uns['merged_sign']
    """

    # blockPrint()

    #  select n_pair*2 closely expressed genes from the data
    genes = get_valid_genes(adata, n_pairs)
    if len(adata.uns["lr"]) > 1:
        raise ValueError("Permutation test only supported for one LR pair scenario.")
    elif type(bg_pairs) == type(None):
        pairs = get_rand_pairs(adata, genes, n_pairs, lrs=adata.uns["lr"])
    else:
        pairs = bg_pairs

    """
    # generate random pairs
    lr1 = adata.uns['lr'][0].split('_')[0]
    lr2 = adata.uns['lr'][0].split('_')[1]
    genes = [item for item in adata.var_names.tolist() if not (item.startswith('MT-') or item.startswith('MT_') or item==lr1 or item==lr2)]
    random.shuffle(genes)
    pairs = [i + '_' + j for i, j in zip(genes[:n_pairs], genes[-n_pairs:])]
    """
    if use_het != None:
        scores = adata.obsm["merged"]
    else:
        scores = adata.obsm[use_lr]

    # for each randomly selected pair, run through cci_rank analysis and keep the scores
    query_pair = adata.uns["lr"]

    # If neighbours not inputted, then compute #
    if type(neighbours) == type(None):
        neighbours = calc_neighbours(adata, distance, index=run_fast)

    if not run_fast and type(background) == type(
        None
    ):  # Run original way if 'fast'=False argument inputted.
        background = []
        for item in pairs:
            adata.uns["lr"] = [item]
            lr(
                adata,
                use_lr=use_lr,
                distance=distance,
                verbose=False,
                neighbours=neighbours,
                **kwargs,
            )
            if use_het != None:
                merge(adata, use_lr=use_lr, use_het=use_het, verbose=False)
                background += adata.obsm["merged"].tolist()
            else:
                background += adata.obsm[use_lr].tolist()
        background = np.array(background)

    elif type(background) == type(None):  # Run fast if background not inputted
        spot_lr1s = get_spot_lrs(adata, pairs, lr_order=True, filter_pairs=False)
        spot_lr2s = get_spot_lrs(adata, pairs, lr_order=False, filter_pairs=False)

        het_vals = (
            np.array([1] * len(adata)) if use_het == None else adata.obsm[use_het]
        )
        background = get_scores(
            spot_lr1s.values, spot_lr2s.values, neighbours, het_vals
        ).ravel()

    # log back the original query
    adata.uns["lr"] = query_pair

    #### Negative Binomial fit
    pvals, pvals_adj, log10_pvals, lr_sign = get_stats(
        scores, background, neg_binom, adj_method
    )

    if use_het != None:
        adata.obsm["merged"] = scores
        adata.obsm["merged_pvalues"] = log10_pvals
        adata.obsm["merged_sign"] = lr_sign

        # enablePrint()
        print(
            "Results of permutation test has been kept in adata.obsm['merged_pvalues']"
        )
        print("Significant merged result has been kept in adata.obsm['merged_sign']")
    else:
        adata.obsm[use_lr] = scores
        adata.obsm["lr_pvalues"] = log10_pvals
        adata.obsm["lr_sign"] = lr_sign  # scores for spots with pval_adj < 0.05

        # enablePrint()
        print("Results of permutation test has been kept in adata.obsm['lr_pvalues']")
        print("Significant merged result has been kept in adata.obsm['lr_sign']")

    # return adata
    return background


def get_stats(
    scores: np.array,
    background: np.array,
    total_bg: int,
    neg_binom: bool = False,
    adj_method: str = "fdr_bh",
    pval_adj_cutoff: float = 0.01,
    return_negbinom_params: bool = False,
):
    """Retrieves valid candidate genes to be used for random gene pairs.
    Parameters
    ----------
    scores: np.array        Per spot scores for a particular LR pair.
    background: np.array    Background distribution for non-zero scores.
    total_bg: int           Total number of background values calculated.
    neg_binom: bool         Whether to use neg-binomial distribution to estimate p-values, NOT appropriate with log1p data, alternative is to use background distribution itself (recommend higher number of n_pairs for this).
    adj_method: str         Parsed to statsmodels.stats.multitest.multipletests for multiple hypothesis testing correction.
    Returns
    -------
    stats: tuple          Per spot pvalues, pvals_adj, log10_pvals_adj, lr_sign (the LR scores for significant spots).
    """
    ##### Negative Binomial fit
    if neg_binom:
        # Need to make full background for fitting !!!
        background = np.array(list(background) + [0] * (total_bg - len(background)))
        pmin, pmax = min(background), max(background)
        background2 = [item - pmin for item in background]
        x = np.linspace(pmin, pmax, 1000)
        res = sm.NegativeBinomial(
            background2, np.ones(len(background2)), loglike_method="nb2"
        ).fit(start_params=[0.1, 0.3], disp=0)
        mu = res.predict()  # use if not constant
        mu = np.exp(res.params[0])
        alpha = res.params[1]
        Q = 0
        size = 1.0 / alpha * mu**Q
        prob = size / (size + mu)

        if return_negbinom_params:  # For testing purposes #
            return size, prob

        # Calculate probability for all spots
        pvals = 1 - scipy.stats.nbinom.cdf(scores - pmin, size, prob)

    else:  ###### Using the actual values to estimate p-values
        pvals = np.zeros((1, len(scores)), dtype=np.float)[0, :]
        nonzero_score_bool = scores > 0
        nonzero_score_indices = np.where(nonzero_score_bool)[0]
        zero_score_indices = np.where(nonzero_score_bool == False)[0]
        pvals[zero_score_indices] = (total_bg - len(background)) / total_bg
        pvals[nonzero_score_indices] = [
            len(np.where(background >= scores[i])[0]) / total_bg
            for i in nonzero_score_indices
        ]

    pvals_adj = multipletests(pvals, method=adj_method)[1]
    log10_pvals_adj = -np.log10(pvals_adj)
    lr_sign = scores * (pvals_adj < pval_adj_cutoff)
    return pvals, pvals_adj, log10_pvals_adj, lr_sign


# @njit(parallel=True)
# def perm_pvals(scores, background):
#     """Determines the p-values based on the actual observed frequency of the \
#         indicated score or greater in the background.
#     """
#     pvals = np.zeros((1, len(scores)), np.float64)[0,:]
#     for i in prange(len(pvals)):


def get_valid_genes(adata: AnnData, n_pairs: int) -> np.array:
    """Retrieves valid candidate genes to be used for random gene pairs.
    Parameters
    ----------
    adata: AnnData          The data object including the cell types to count
    n_pairs: int            The number of random pairs will generate elsewhere.
    Returns
    -------
    genes: np.array          List of genes which could be valid pairs.
    """
    genes = np.array(
        [
            item
            for item in adata.var_names.tolist()
            if not (item.startswith("MT-") or item.startswith("MT_"))
        ]
    )
    if n_pairs >= len(genes) / 2:
        raise ValueError(
            "Too many genes pairs selected, please reduce to a smaller number."
        )
    return genes


def get_rand_pairs(
    adata: AnnData,
    genes: np.array,
    n_pairs: int,
    lrs: list = None,
    im: int = None,
):
    """Gets equivalent random gene pairs for the inputted lr pair.
    Parameters
    ----------
    adata: AnnData          The data object including the cell types to count
    lr: int            The lr pair string to get equivalent random pairs for (e.g. 'L_R')
    genes: np.array           Candidate genes to use as pairs.
    n_pairs: int             Number of random pairs to generate.
    Returns
    -------
    pairs: list          List of random gene pairs with equivalent mean expression (e.g. ['L_R'])
    """
    lr_genes = [lr.split("_")[0] for lr in lrs]
    lr_genes += [lr.split("_")[1] for lr in lrs]

    # get the position of the median of the means between the two genes
    means_ordered, genes_ordered = get_ordered(adata, genes)
    if type(im) == type(None):  # Single background per lr pair mode
        l, r = lrs[0].split("_")
        im = get_median_index(l, r, means_ordered.values, genes_ordered)

    # get n_pair genes sorted by distance to im
    selected = (
        abs(means_ordered - means_ordered[im])
        .sort_values()
        .drop(lr_genes)[: n_pairs * 2]
        .index.tolist()
    )
    selected = selected[0 : n_pairs * 2]
    adata.uns["selected"] = selected
    # form gene pairs from selected randomly
    random.shuffle(selected)
    pairs = [i + "_" + j for i, j in zip(selected[:n_pairs], selected[-n_pairs:])]

    return pairs


def get_ordered(adata, genes):
    means_ordered = adata.to_df()[genes].mean().sort_values()
    genes_ordered = means_ordered.index.values
    return means_ordered, genes_ordered


def get_median_index(l, r, means_ordered, genes_ordered):
    """ "Retrieves the index of the gene with a mean expression between the two genes in the lr pair.
    Parameters
    ----------
    X: np.ndarray          Spot*Gene expression.
    l: str                 Ligand gene.
    r: str                 Receptor gene.
    genes: np.array        Candidate genes to use as pairs.
    Returns
    -------
    pairs: list          List of random gene pairs with equivalent mean expression (e.g. ['L_R'])
    """
    # sort the mean of each gene expression
    i1 = np.where(genes_ordered == l)[0][0]
    i2 = np.where(genes_ordered == r)[0][0]
    if means_ordered[i1] > means_ordered[i2]:
        it = i1
        i1 = i2
        i2 = it

    im = np.argmin(np.abs(means_ordered - np.median(means_ordered[i1:i2])))
    return im
    # means = adata.to_df()[genes].mean().sort_values()
    # lr1 = lr[0].split("_")[0]
    # lr2 = lr[0].split("_")[1]
    # i1, i2 = means.index.get_loc(lr1), means.index.get_loc(lr2)
    # if means[lr1] > means[lr2]:
    #     it = i1
    #     i1 = i2
    #     i2 = it
    #
    # # get the position of the median of the means between the two genes
    # im = np.argmin(abs(means.values - means.iloc[i1:i2]))


# Disable printing
def blockPrint():
    sys.stdout = open(os.devnull, "w")


# Restore printing
def enablePrint():
    sys.stdout = sys.__stdout__



doc_spatial_base_plot = """\
adata
    Annotated data matrix.
title
    Title name of the figure.
figsize
    Figure size with the format (width,height).
cmap
    Color map to use for continous variables or discretes variables (e.g. viridis, Set1,...).
use_label
    Key for the label use in `adata.obs` (e.g. `leiden`, `louvain`,...).
list_clusters
    A set of cluster to be displayed in the figure (e.g. [0,1,2,3]).
ax
    A matplotlib axes object.
show_plot
    Option to display the figure.
show_image
    Option to display the H&E image.
show_color_bar
    Option to display color bar.
crop
    Option to crop the figure based on the spot locations.
margin
    Margin to crop.
size
    Spot size to display in figure.
image_alpha
    Opacity of H&E image.
cell_alpha
    Opacity of spots/cells.
use_raw
    Option to use `adata.raw` data.
fname
    Output path to the output if user want to save the figure.
dpi
    Dots per inch values for the output.
"""

doc_gene_plot = """\
gene_symbols
    Single gene (str) or multiple genes (list) that user wants to display. It should be available in `adata.var_names`.
threshold
    Threshold to display genes in the figure.
method
    Method to combine multiple genes:
    `'CumSum'` is cummulative sum of genes expression values,
    `'NaiveMean'` is the mean of the genes expression values.
contour
    Option to show the contour plot.
step_size
    Determines the number and positions of the contour lines / regions.
"""

doc_cluster_plot = """\
show_subcluster
    Display the subcluster in the figure.
show_cluster_labels
    Display the labels of clusters.
show_trajectories
    Display the spatial trajectory analysis results.
reverse
    Reverse the direction of spatial trajectories.
show_node
    Show node of PAGA graph mapping in spatial.
threshold_spots
    The number of spots threshold for not display the subcluster labels
text_box_size
    The font size in the box of labels.
color_bar_size
    The size of color bar.
bbox_to_anchor
    Set the position of box of color bar. Default is `(1,1)`
"""

doc_lr_plot = """\
adata
    AnnData object with run st.tl.cci_rank.run performed on.
lr
    Ligand receptor paid (in format L_R)
min_expr
    Minimum expression for gene to be considered expressed.
sig_spots
    Whether to filter to significant spots or not.
use_label
    Label to use for the inner points, can be in adata.obs or in the lr stats of adata.uns['per_lr_results'][lr].columns
use_mix
    The deconvolution/label_transfer results to use for visualising pie charts in the inner point, not currently implimented.
outer_mode
    Either 'binary', 'continuous', or None; controls how ligand-receptor expression shown (or not shown).
l_cmap
    matplotlib cmap controlling ligand continous expression.
r_cmap
    matplotlib cmap controlling receptor continuous expression.
lr_cmap
    matplotlib cmap controlling the ligand receptor binary expression, but have atleast 4 colours.
inner_cmap
    matplotlib cmap controlling the inner point colours.
inner_size_prop
    multiplier which controls size of inner points.
middle_size_prop
    Multiplier which controls size of middle point (only relevant when outer_mode='continuous')
outer_size_prop
    Multiplier which controls size of the outter point.
pt_scale
    Multiplier which scales overall point size of all points plotted.
title
    Title of the plot.
show_image
    Whether to show the background H&E or not.
kwargs
    Extra arguments parsed to the other plotting functions such as gene_plot, cluster_plot, &/or het_plot.
"""

doc_het_plot = """\
use_het
    Single gene (str) or multiple genes (list) that user wants to display. It should be available in `adata.var_names`.
contour
    Option to show the contour plot.
step_size
    Determines the number and positions of the contour lines / regions.
vmin
    Lower end of scale bar.
vmax
    Upper end of scale bar.
"""

doc_subcluster_plot = """\
cluster
    Choose cluster to plot the sub-clusters.
text_box_size
    The font size in the box of labels.
bbox_to_anchor
    Set the position of box of color bar. Default is `(1,1)`
"""



import matplotlib
from matplotlib import pyplot as plt

from typing import Optional, Union
from anndata import AnnData


def plot_mask(
    adata: AnnData,
    library_id: str = None,
    show_spot: bool = True,
    spot_alpha: float = 1.0,
    cmap: str = "vega_20_scanpy",
    tissue_alpha: float = 1.0,
    mask_alpha: float = 0.5,
    spot_size: Union[float, int] = 6.5,
    show_legend: bool = True,
    name: str = "mask_plot",
    dpi: int = 150,
    output: str = None,
    show_axis: bool = False,
    show_plot: bool = True,
) -> Optional[AnnData]:
    """\
    mask plot for sptial transcriptomics data.

    Parameters
    ----------
    adata
        Annotated data matrix.
    library_id
        Library id stored in AnnData.
    show_spot
        Show spot or not
    spot_alpha
        Opacity of the spot.
    cmap
        Color map to use.
    tissue_alpha
        Opacity of the tissue.
    mask_alpha
        Opacity of the mask.
    spot_size
        Size of the spot.
    show_axis
        Show axis or not.
    show_legend
        Show legend or not.
    name
        Name of the output figure file.
    dpi
        DPI of the output figure.
    output
        Save the figure as file or not.
    show_plot
        Show plot or not
    Returns
    -------
    Nothing
    """
    from scanpy.plotting import palettes
    from stlearn.plotting import palettes_st

    if cmap == "vega_10_scanpy":
        cmap = palettes.vega_10_scanpy
    elif cmap == "vega_20_scanpy":
        cmap = palettes.vega_20_scanpy
    elif cmap == "default_102":
        cmap = palettes.default_102
    elif cmap == "default_28":
        cmap = palettes.default_28
    elif cmap == "jana_40":
        cmap = palettes_st.jana_40
    elif cmap == "default":
        cmap = palettes_st.default
    else:
        raise ValueError(
            "We only support vega_10_scanpy, vega_20_scanpy, default_28, default_102"
        )

    cmaps = matplotlib.colors.LinearSegmentedColormap.from_list("", cmap)

    cmap_ = plt.cm.get_cmap(cmaps)

    plt.rcParams["figure.dpi"] = dpi

    # Option for turning off showing figure
    plt.ioff()

    # Initialize matplotlib
    fig, a = plt.subplots()

    x_min, x_max = adata.obs["imagecol"].min(), adata.obs["imagecol"].max()
    y_min, y_max = adata.obs["imagerow"].min(), adata.obs["imagerow"].max()
    add_x = (x_max - x_min) // 20
    add_y = (y_max - y_min) // 20

    plt.xlim([x_min - add_x, x_max + add_x])
    plt.ylim([y_max + add_y, y_min - add_y])
    key = "mask_annotation"
    if show_spot:
        for i, cluster in enumerate(adata.obs.groupby(key)):
            if cluster[1][key + "_code"][0] == -1:
                # Plot scatter plot based on pixel of spots
                _ = a.scatter(
                    cluster[1]["imagecol"],
                    cluster[1]["imagerow"],
                    edgecolor="none",
                    alpha=spot_alpha,
                    s=spot_size,
                    marker="o",
                    c=[(1.0, 1.0, 1.0, 1.0)],
                )
            else:
                # Plot scatter plot based on pixel of spots
                _ = a.scatter(
                    cluster[1]["imagecol"],
                    cluster[1]["imagerow"],
                    edgecolor="none",
                    alpha=spot_alpha,
                    s=spot_size,
                    marker="o",
                    c=[cmap_(int(cluster[1][key + "_code"][0]) / (len(cmap) - 1))],
                )

    if show_legend:
        from matplotlib.patches import Patch

        legend_elements = []
        for index, row in adata.obs.groupby(key).first().reset_index().iterrows():
            if row[key + "_code"] == -1:
                legend_elements.append(
                    Patch(color=(1.0, 1.0, 1.0, 1.0), label=row[key])
                )
            else:
                legend_elements.append(
                    Patch(
                        color=cmap_(int(row[key + "_code"]) / (len(cmap) - 1)),
                        label=row[key],
                    )
                )

        a.legend(
            handles=legend_elements,
            loc="center left",
            bbox_to_anchor=(1, 0.5),
            prop=dict(size=8),
        )

    if library_id is None:
        library_id = list(adata.uns["spatial"].keys())[0]

    image = adata.uns["spatial"][library_id]["images"][
        adata.uns["spatial"][library_id]["use_quality"]
    ]

    if not show_axis:
        a.axis("off")
    # Overlay the tissue image
    a.imshow(
        image,
        alpha=tissue_alpha,
        zorder=-1,
    )
    a.imshow(
        adata.uns[key],
        alpha=mask_alpha,
        zorder=-1,
    )

    if output is not None:
        fig.savefig(output + "/" + name, dpi=dpi, bbox_inches="tight", pad_inches=0)

    if show_plot == True:
        plt.show()



import numpy as np
import pandas as pd
import networkx as nx
from .global_level import global_level
from .local_level import local_level
from .utils import lambda_dist, resistance_distance
from tqdm import tqdm


def weight_optimizing_global(
    adata,
    use_label=None,
    list_clusters=None,
    step=0.01,
    k=10,
    use_rep="X_pca",
    n_dims=40,
):
    # Screening PTS graph
    print("Screening PTS global graph...")
    Gs = []
    j = 0

    with tqdm(
        total=int(1 / step + 1),
        desc="Screening",
        bar_format="{l_bar}{bar} [ time left: {remaining} ]",
    ) as pbar:
        for i in range(0, int(1 / step + 1)):

            Gs.append(
                nx.to_scipy_sparse_array(
                    global_level(
                        adata,
                        use_label=use_label,
                        list_clusters=list_clusters,
                        use_rep=use_rep,
                        n_dims=n_dims,
                        w=round(j, 2),
                        return_graph=True,
                        verbose=False,
                    )
                )
            )

            j = j + step
            pbar.update(1)

    # Calculate the graph dissimilarity using Laplacian matrix
    print("Calculate the graph dissimilarity using Laplacian matrix...")
    result = []
    a1_list = []
    a2_list = []
    indx = []
    w = 0
    k = len(
        adata.obs[adata.obs[use_label].isin(list_clusters)][
            "sub_cluster_labels"
        ].unique()
    )
    with tqdm(
        total=int(1 / step - 1),
        desc="Calculating",
        bar_format="{l_bar}{bar} [ time left: {remaining} ]",
    ) as pbar:
        for i in range(1, int(1 / step)):
            w += step
            a1 = lambda_dist(Gs[i], Gs[0], k=k)
            a2 = lambda_dist(Gs[i], Gs[-1], k=k)
            a1_list.append(a1)
            a2_list.append(a2)
            indx.append(w)
            result.append(np.absolute(1 - a1 / a2))
            pbar.update(1)

    screening_result = pd.DataFrame(
        {"w": indx, "A1": a1_list, "A2": a2_list, "Dissmilarity_Score": result}
    )

    adata.uns["screening_result_global"] = screening_result

    def NormalizeData(data):
        return (data - np.min(data)) / (np.max(data) - np.min(data))

    result = NormalizeData(result)

    try:
        optimized_ind = np.where(result == np.amin(result))[0][0]
        opt_w = round(indx[optimized_ind], 2)
        print("The optimized weighting is:", str(opt_w))
        return opt_w
    except:
        print("The optimized weighting is: 0.5")
        return 0.5


def weight_optimizing_local(adata, use_label=None, cluster=None, step=0.01):
    # Screening PTS graph
    print("Screening PTS local graph...")
    Gs = []
    j = 0
    with tqdm(
        total=int(1 / step + 1),
        desc="Screening",
        bar_format="{l_bar}{bar} [ time left: {remaining} ]",
    ) as pbar:
        for i in range(0, int(1 / step + 1)):

            Gs.append(
                local_level(
                    adata,
                    use_label=use_label,
                    cluster=cluster,
                    w=round(j, 2),
                    verbose=False,
                    return_matrix=True,
                )
            )

            j = j + step
            pbar.update(1)

    # Calculate the graph dissimilarity using Laplacian matrix
    print("Calculate the graph dissimilarity using Resistance distance...")
    result = []
    a1_list = []
    a2_list = []
    indx = []
    w = 0

    with tqdm(
        total=int(1 / step - 1),
        desc="Calculating",
        bar_format="{l_bar}{bar} [ time left: {remaining} ]",
    ) as pbar:
        for i in range(1, int(1 / step)):
            w += step
            a1 = resistance_distance(Gs[i], Gs[0])
            a2 = resistance_distance(Gs[i], Gs[-1])
            a1_list.append(a1)
            a2_list.append(a2)
            indx.append(w)
            result.append(np.absolute(1 - a1 / a2))
            pbar.update(1)

    screening_result = pd.DataFrame(
        {"w": indx, "A1": a1_list, "A2": a2_list, "Dissmilarity_Score": result}
    )

    adata.uns["screening_result_local"] = screening_result

    def NormalizeData(data):
        return (data - np.min(data)) / (np.max(data) - np.min(data))

    result = NormalizeData(result)

    optimized_ind = np.where(result == np.amin(result))[0][0]
    opt_w = round(indx[optimized_ind], 2)
    print("The optimized weighting is:", str(opt_w))

    return opt_w



from anndata import AnnData
from typing import Optional, Union
import matplotlib.pyplot as plt
import scanpy as sc
import numpy as np


def check_trajectory(
    adata: AnnData,
    library_id: str = None,
    use_label: str = "louvain",
    basis: str = "umap",
    pseudotime_key: str = "dpt_pseudotime",
    trajectory: list = None,
    figsize=(10, 4),
    size_umap: int = 50,
    size_spatial: int = 1.5,
    img_key: str = "hires",
) -> Optional[AnnData]:
    trajectory = np.array(trajectory).astype(int)
    assert (
        trajectory in adata.uns["available_paths"].values()
    ), "Please choose the right path!"
    trajectory = trajectory.astype(str)
    assert (
        pseudotime_key in adata.obs.columns
    ), "Please run the pseudotime or choose the right one!"
    assert (
        use_label in adata.obs.columns
    ), "Please run the cluster or choose the right label!"
    assert basis in adata.obsm, (
        "Please run the " + basis + "before you check the trajectory!"
    )
    if library_id is None:
        library_id = list(adata.uns["spatial"].keys())[0]

    adata.obsm["X_spatial"] = adata.obs[["imagecol", "imagerow"]].values

    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=figsize)

    ax1 = sc.pl.umap(adata, size=size_umap, show=False, ax=ax1)
    sc.pl.umap(
        adata[adata.obs[use_label].isin(trajectory)],
        size=size_umap,
        color=pseudotime_key,
        ax=ax1,
        show=False,
        frameon=False,
    )

    ax2 = sc.pl.scatter(
        adata,
        size=25,
        show=False,
        basis="spatial",
        ax=ax2,
    )
    sc.pl.spatial(
        adata[adata.obs[use_label].isin(trajectory)],
        size=size_spatial,
        ax=ax2,
        color=pseudotime_key,
        legend_loc="none",
        basis="spatial",
        frameon=False,
        show=False,
    )

    im = ax2.imshow(
        adata.uns["spatial"][library_id]["images"][img_key], alpha=0, zorder=-1
    )

    plt.show()

    del adata.obsm["X_spatial"]



"""Top-level package for stLearn."""

__author__ = """Genomics and Machine Learning lab"""
__email__ = "duy.pham@uq.edu.au"
__version__ = "0.4.9"


from . import add
from . import pp
from . import em
from . import tl
from . import pl
from . import spatial
from . import datasets

# Wrapper

from .wrapper.read import ReadSlideSeq
from .wrapper.read import Read10X
from .wrapper.read import ReadOldST
from .wrapper.read import ReadMERFISH
from .wrapper.read import ReadSeqFish
from .wrapper.read import ReadXenium
from .wrapper.read import create_stlearn

from ._settings import settings
from .wrapper.convert_scanpy import convert_scanpy
from .wrapper.concatenate_spatial_adata import concatenate_spatial_adata

# from . import cli



"""Logging and Profiling
"""
import logging
from functools import update_wrapper, partial
from logging import CRITICAL, ERROR, WARNING, INFO, DEBUG, NOTSET
from datetime import datetime, timedelta, timezone
from typing import Optional

import anndata.logging


HINT = (INFO + DEBUG) // 2
logging.addLevelName(HINT, "HINT")


class _RootLogger(logging.RootLogger):
    def __init__(self, level):
        super().__init__(level)
        self.propagate = False
        _RootLogger.manager = logging.Manager(self)

    def log(
        self,
        level: int,
        msg: str,
        *,
        extra: Optional[dict] = None,
        time: datetime = None,
        deep: Optional[str] = None,
    ) -> datetime:
        from . import settings

        now = datetime.now(timezone.utc)
        time_passed: timedelta = None if time is None else now - time
        extra = {
            **(extra or {}),
            "deep": deep if settings.verbosity.level < level else None,
            "time_passed": time_passed,
        }
        super().log(level, msg, extra=extra)
        return now

    def critical(self, msg, *, time=None, deep=None, extra=None) -> datetime:
        return self.log(CRITICAL, msg, time=time, deep=deep, extra=extra)

    def error(self, msg, *, time=None, deep=None, extra=None) -> datetime:
        return self.log(ERROR, msg, time=time, deep=deep, extra=extra)

    def warning(self, msg, *, time=None, deep=None, extra=None) -> datetime:
        return self.log(WARNING, msg, time=time, deep=deep, extra=extra)

    def info(self, msg, *, time=None, deep=None, extra=None) -> datetime:
        return self.log(INFO, msg, time=time, deep=deep, extra=extra)

    def hint(self, msg, *, time=None, deep=None, extra=None) -> datetime:
        return self.log(HINT, msg, time=time, deep=deep, extra=extra)

    def debug(self, msg, *, time=None, deep=None, extra=None) -> datetime:
        return self.log(DEBUG, msg, time=time, deep=deep, extra=extra)


def _set_log_file(settings):
    file = settings.logfile
    name = settings.logpath
    root = settings._root_logger
    h = logging.StreamHandler(file) if name is None else logging.FileHandler(name)
    h.setFormatter(_LogFormatter())
    h.setLevel(root.level)
    if len(root.handlers) == 1:
        root.removeHandler(root.handlers[0])
    elif len(root.handlers) > 1:
        raise RuntimeError("Scanpy’s root logger somehow got more than one handler")
    root.addHandler(h)


def _set_log_level(settings, level: int):
    root = settings._root_logger
    root.setLevel(level)
    (h,) = root.handlers  # may only be 1
    h.setLevel(level)


class _LogFormatter(logging.Formatter):
    def __init__(
        self, fmt="{levelname}: {message}", datefmt="%Y-%m-%d %H:%M", style="{"
    ):
        super().__init__(fmt, datefmt, style)

    def format(self, record: logging.LogRecord):
        format_orig = self._style._fmt
        if record.levelno == INFO:
            self._style._fmt = "{message}"
        elif record.levelno == HINT:
            self._style._fmt = "--> {message}"
        elif record.levelno == DEBUG:
            self._style._fmt = "    {message}"
        if record.time_passed:
            # strip microseconds
            if record.time_passed.microseconds:
                record.time_passed = timedelta(
                    seconds=int(record.time_passed.total_seconds())
                )
            if "{time_passed}" in record.msg:
                record.msg = record.msg.replace(
                    "{time_passed}", str(record.time_passed)
                )
            else:
                self._style._fmt += " ({time_passed})"
        if record.deep:
            record.msg = f"{record.msg}: {record.deep}"
        result = logging.Formatter.format(self, record)
        self._style._fmt = format_orig
        return result


print_memory_usage = anndata.logging.print_memory_usage
get_memory_usage = anndata.logging.get_memory_usage


_DEPENDENCIES_NUMERICS = [
    "anndata",  # anndata actually shouldn't, but as long as it's in development
    "umap",
    "numpy",
    "scipy",
    "pandas",
    ("sklearn", "scikit-learn"),
    "statsmodels",
    ("igraph", "python-igraph"),
    "louvain",
]


_DEPENDENCIES_PLOTTING = ["matplotlib", "seaborn"]


def _versions_dependencies(dependencies):
    # this is not the same as the requirements!
    for mod in dependencies:
        mod_name, dist_name = mod if isinstance(mod, tuple) else (mod, mod)
        try:
            imp = __import__(mod_name)
            yield dist_name, imp.__version__
        except (ImportError, AttributeError):
            pass


def print_versions():
    """\
    Versions that might influence the numerical results.

    Matplotlib and Seaborn are excluded from this.
    """
    from ._settings import settings

    modules = ["scanpy"] + _DEPENDENCIES_NUMERICS
    print(
        " ".join(f"{mod}=={ver}" for mod, ver in _versions_dependencies(modules)),
        file=settings.logfile,
    )


def print_version_and_date():
    """\
    Useful for starting a notebook so you see when you started working.
    """
    from . import __version__
    from ._settings import settings

    print(
        f"Running Scanpy {__version__}, " f"on {datetime.now():%Y-%m-%d %H:%M}.",
        file=settings.logfile,
    )


def _copy_docs_and_signature(fn):
    return partial(update_wrapper, wrapped=fn, assigned=["__doc__", "__annotations__"])


def error(
    msg: str,
    *,
    time: datetime = None,
    deep: Optional[str] = None,
    extra: Optional[dict] = None,
) -> datetime:
    """\
    Log message with specific level and return current time.

    Parameters
    ==========
    msg
        Message to display.
    time
        A time in the past. If this is passed, the time difference from then
        to now is appended to `msg` as ` (HH:MM:SS)`.
        If `msg` contains `{time_passed}`, the time difference is instead
        inserted at that position.
    deep
        If the current verbosity is higher than the log function’s level,
        this gets displayed as well
    extra
        Additional values you can specify in `msg` like `{time_passed}`.
    """
    from ._settings import settings

    return settings._root_logger.error(msg, time=time, deep=deep, extra=extra)


@_copy_docs_and_signature(error)
def warning(msg, *, time=None, deep=None, extra=None) -> datetime:
    from ._settings import settings

    return settings._root_logger.warning(msg, time=time, deep=deep, extra=extra)


@_copy_docs_and_signature(error)
def info(msg, *, time=None, deep=None, extra=None) -> datetime:
    from ._settings import settings

    return settings._root_logger.info(msg, time=time, deep=deep, extra=extra)


@_copy_docs_and_signature(error)
def hint(msg, *, time=None, deep=None, extra=None) -> datetime:
    from ._settings import settings

    return settings._root_logger.hint(msg, time=time, deep=deep, extra=extra)


@_copy_docs_and_signature(error)
def debug(msg, *, time=None, deep=None, extra=None) -> datetime:
    from ._settings import settings

    return settings._root_logger.debug(msg, time=time, deep=deep, extra=extra)



from typing import Optional, Union
from anndata import AnnData
import pandas as pd
import numpy as np
from pathlib import Path


def add_deconvolution(
    adata: AnnData,
    annotation_path: Union[Path, str],
    copy: bool = False,
) -> Optional[AnnData]:

    """\
    Adding label transfered from Seurat

    Parameters
    ----------
    adata
        Annotated data matrix.
    annotation_path
        Path of the output of label transfer result by Seurat
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    Depending on `copy`, returns or updates `adata` with the following fields.
    **[cluster method name]_anno** : `adata.obs` field
        The annotation of cluster results.

    """
    label = pd.read_csv(annotation_path, index_col=0)
    label = label[adata.obs_names]

    adata.obsm["deconvolution"] = label[adata.obs.index].T



"""
Tests that everything is installed correctly.
"""

import unittest


class TestCCI(unittest.TestCase):
    """Tests for `stlearn` importability, i.e. correct installation."""

    def test_SME(self):
        import stlearn.spatials.SME.normalize as sme_normalise

    def test_cci(self):
        """Tests CCI can be imported."""
        import stlearn.tools.microenv.cci.analysis as an



# Purpose of this script is to write the functions that help facilitate
# subsetting of the data depending on the users input
import numpy


def printOut(text, fileName="stdout.txt", close=True, file=None):
    """Prints to the specified file name. Used for debugging.
    If close is Fale, returns open file.
    """

    if type(file) == type(None):
        file = open(fileName, "w")

    print(text, file=file)

    if close:
        file.close()
    else:
        return file


def filterOptions(metaDataSets, options):
    """Returns options that overlap with keys in metaDataSets dictionary"""
    if type(options) == type(None):
        options = list(metaDataSets.keys())
    else:
        options = [option for option in options if option in metaDataSets.keys()]

    return options


def addChoices(metaDataSets, options, elementValues):
    """Helper function which generates choices for SelectMultiField"""
    for option in options:
        choices = [(optioni, optioni) for optioni in metaDataSets[option]]
        elementValues.append(choices)


# TODO update this so has 'options' as input
def subsetSCA(sca, subsetForm):
    """Subsets the SCA based on the selected fields and the inputted genes."""

    # Getting the attached fields from the form which refer subset options #
    options = filterOptions(sca.metaDataSets, subsetForm.elements)

    # Subsetting based on selection #
    conditionSelection = {}  # selection dictionary
    for i, option in enumerate(options):
        selected = getattr(subsetForm, option).data
        if len(selected) != 0:
            conditionSelection[option] = selected

    # Subsetting based on conditions #
    if len(conditionSelection) != 0:
        sca = sca.createConditionSubset("subset", conditionSelection)

    # Subsetting based on inputted genes #
    geneList = getattr(subsetForm, "Select Cells Expressing Gene/s").data.split(",")
    if geneList != [""]:
        # Filter to just the genes which express all of the inputted genes #
        sca = sca.createGeneExprsSubset(
            "subset", genesToFilter=geneList, cutoff=0, keep=True, useOr=False
        )

    return sca, conditionSelection, geneList



# -*- coding: utf-8 -*-
"""Helper utilities and decorators."""
from flask import flash
import matplotlib.pyplot as plt


def flash_errors(form, category="warning"):
    """Flash all errors for a form."""
    for field, errors in form.errors.items():
        for error in errors:
            flash(getattr(form, field).label.text + " - " + error + ", category")


def get_all_paths(adata):

    import networkx as nx

    G = nx.from_numpy_array(adata.uns["paga"]["connectivities_tree"].toarray())
    mapping = {int(k): v for k, v in zip(G.nodes, adata.obs.clusters.cat.categories)}
    G = nx.relabel_nodes(G, mapping)

    all_paths = []
    for source in G.nodes:
        for target in G.nodes:
            paths = nx.all_simple_paths(G, source=source, target=target)
            for path in paths:
                all_paths.append(path)

    import numpy as np

    all_paths = list(map(lambda x: " - ".join(np.array(x).astype(str)), all_paths))

    return all_paths



from typing import Optional, Union
from anndata import AnnData
import pandas as pd


def stack_3d_plot(
    adata: AnnData,
    slides,
    cmap="viridis",
    slide_col="sample_id",
    use_label=None,
    gene_symbol=None,
) -> Optional[AnnData]:

    """\
    Clustering plot for sptial transcriptomics data. Also it has a function to display trajectory inference.

    Parameters
    ----------
    adata
        Annotated data matrix.
    slides
        A list of slide id
    cmap
        Color map
    use_label
        Choose label to plot (priotize)
    gene_symbol
        Choose gene symbol to plot
    width
        Witdh of the plot
    height
        Height of the plot
    Returns
    -------
    Nothing
    """
    try:
        import plotly.express as px
    except ModuleNotFoundError:
        raise ModuleNotFoundError("Please install plotly by `pip install plotly`")

    assert (
        slide_col in adata.obs.columns
    ), "Please provide the right column for slide_id!"

    list_df = []
    for i, slide in enumerate(slides):
        tmp = data.obs[data.obs[slide_col] == slide][["imagecol", "imagerow"]]
        tmp["sample_id"] = slide
        tmp["z-dimension"] = i
        list_df.append(tmp)

    df = pd.concat(list_df)

    if use_label != None:
        assert use_label in adata.obs.columns, "Please use the right `use_label`"
        df[use_label] = adata[df.index].obs[use_label].values

        fig = px.scatter_3d(
            df,
            x="imagecol",
            y="imagerow",
            z="z-dimension",
            color=use_label,
            width=width,
            height=height,
            color_continuous_scale=cmap,
        )
        fig.show()

    else:
        assert gene_symbol in adata.var_names, "Please use the right `gene_symbol`"
        df[gene_symbol] = adata[df.index][:, gene_symbol].X

        fig = px.scatter_3d(
            df,
            x="imagecol",
            y="imagerow",
            z="z-dimension",
            color=gene_symbol,
            width=width,
            height=height,
            color_continuous_scale=cmap,
        )
        fig.show()



from matplotlib import pyplot as plt
from PIL import Image
import pandas as pd
import matplotlib
import numpy as np
import networkx as nx

from typing import Optional, Union, Mapping  # Special
from typing import Sequence, Iterable  # ABCs
from typing import Tuple  # Classes

from anndata import AnnData
import warnings

from stlearn.plotting.classes import ClusterPlot
from stlearn.plotting.classes_bokeh import BokehClusterPlot
from stlearn.plotting._docs import doc_spatial_base_plot, doc_cluster_plot
from stlearn.utils import _AxesSubplot, Axes, _docs_params

from bokeh.io import push_notebook, output_notebook
from bokeh.plotting import show


@_docs_params(spatial_base_plot=doc_spatial_base_plot, cluster_plot=doc_cluster_plot)
def cluster_plot(
    adata: AnnData,
    # plotting param
    title: Optional["str"] = None,
    figsize: Optional[Tuple[float, float]] = None,
    cmap: Optional[str] = "default",
    use_label: Optional[str] = None,
    list_clusters: Optional[list] = None,
    ax: Optional[matplotlib.axes._subplots.Axes] = None,
    fig: Optional[matplotlib.figure.Figure] = None,
    show_plot: Optional[bool] = True,
    show_axis: Optional[bool] = False,
    show_image: Optional[bool] = True,
    show_color_bar: Optional[bool] = True,
    zoom_coord: Optional[float] = None,
    crop: Optional[bool] = True,
    margin: Optional[bool] = 100,
    size: Optional[float] = 5,
    image_alpha: Optional[float] = 1.0,
    cell_alpha: Optional[float] = 1.0,
    fname: Optional[str] = None,
    dpi: Optional[int] = 120,
    # cluster plot param
    show_subcluster: Optional[bool] = False,
    show_cluster_labels: Optional[bool] = False,
    show_trajectories: Optional[bool] = False,
    reverse: Optional[bool] = False,
    show_node: Optional[bool] = False,
    threshold_spots: Optional[int] = 5,
    text_box_size: Optional[float] = 5,
    color_bar_size: Optional[float] = 10,
    bbox_to_anchor: Optional[Tuple[float, float]] = (1, 1),
    # trajectory
    trajectory_node_size: Optional[int] = 10,
    trajectory_alpha: Optional[float] = 1.0,
    trajectory_width: Optional[float] = 2.5,
    trajectory_edge_color: Optional[str] = "#f4efd3",
    trajectory_arrowsize: Optional[int] = 17,
) -> Optional[AnnData]:

    """\
    Allows the visualization of a cluster results as the discretes values
    of dot points in the Spatial transcriptomics array. We also support to
    visualize the spatial trajectory results


    Parameters
    -------------------------------------
    {spatial_base_plot}
    {cluster_plot}

    Examples
    -------------------------------------
    >>> import stlearn as st
    >>> adata = st.datasets.example_bcba()
    >>> label = "louvain"
    >>> st.pl.cluster_plot(adata, use_label = label, show_trajectories = True)

    """

    assert use_label != None, "Please select `use_label` parameter"

    ClusterPlot(
        adata,
        title=title,
        figsize=figsize,
        cmap=cmap,
        use_label=use_label,
        list_clusters=list_clusters,
        ax=ax,
        fig=fig,
        show_plot=show_plot,
        show_axis=show_axis,
        show_image=show_image,
        show_color_bar=show_color_bar,
        zoom_coord=zoom_coord,
        crop=crop,
        margin=margin,
        size=size,
        image_alpha=image_alpha,
        cell_alpha=cell_alpha,
        fname=fname,
        dpi=dpi,
        show_subcluster=show_subcluster,
        show_cluster_labels=show_cluster_labels,
        show_trajectories=show_trajectories,
        reverse=reverse,
        show_node=show_node,
        threshold_spots=threshold_spots,
        text_box_size=text_box_size,
        color_bar_size=color_bar_size,
        bbox_to_anchor=bbox_to_anchor,
        trajectory_node_size=trajectory_node_size,
        trajectory_alpha=trajectory_alpha,
        trajectory_width=trajectory_width,
        trajectory_edge_color=trajectory_edge_color,
        trajectory_arrowsize=trajectory_arrowsize,
    )


def cluster_plot_interactive(
    adata: AnnData,
):

    bokeh_object = BokehClusterPlot(adata)
    output_notebook()
    show(bokeh_object.app, notebook_handle=True)



from typing import Optional, Union
from anndata import AnnData
import pandas as pd
import numpy as np
import stlearn
from pathlib import Path
from natsort import natsorted


def add_loupe_clusters(
    adata: AnnData,
    loupe_path: Union[Path, str],
    key_add: str = "multiplex",
    copy: bool = False,
) -> Optional[AnnData]:

    """\
    Adding label transfered from Seurat

    Parameters
    ----------
    adata
        Annotated data matrix.
    annotation_path
        Path of the output of label transfer result by Seurat
    use_label
        Choosing cluster type.
    threshold
        Quantile threshold of label
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    Depending on `copy`, returns or updates `adata` with the following fields.
    **[cluster method name]_anno** : `adata.obs` field
        The annotation of cluster results.

    """
    label = pd.read_csv(loupe_path)

    adata.obs[key_add] = pd.Categorical(
        values=np.array(label[key_add]).astype("U"),
        categories=natsorted(label[key_add].unique().astype("U")),
    )



import os, sys, subprocess

sys.path.append(os.path.dirname(__file__))

try:
    import flask
except ImportError:
    subprocess.call(
        "pip install -r " + os.path.dirname(__file__) + "//requirements.txt", shell=True
    )

from flask import (
    Flask,
    render_template,
    request,
    flash,
    url_for,
    redirect,
    session,
    send_file,
)
from bokeh.embed import components
from bokeh.plotting import figure
from bokeh.resources import INLINE
from werkzeug.utils import secure_filename
import tempfile
import traceback

import tempfile
import shutil

import stlearn
import scanpy
import numpy
import numpy as np

import asyncio
from bokeh.server.server import BaseServer
from bokeh.server.tornado import BokehTornado
from tornado.httpserver import HTTPServer
from tornado.ioloop import IOLoop
from bokeh.application import Application
from bokeh.application.handlers import FunctionHandler
from bokeh.server.server import Server
from bokeh.embed import server_document

from bokeh.layouts import column, row

# Functions related to processing the forms.
from source.forms import views  # for changing data in response to input

# Global variables.

global adata  # Storing the data
adata = None
global step_log  # Keeps track of what step we're up to (performed preprocessing?)
step_log = {
    "uploaded": [False, "Upload file"],
    "preprocessed": [False, "Preprocessing"],
    "clustering": [False, "Clustering"],
    "psts": [False, "Spatial trajectory"],
    "dea": [False, "DEA"],
    "lr": [False, "Ligand-receptor analysis"],
    "cci": [False, "CCI"],
    # _params suffix important for templates/progress.html
    "preprocessed_params": {},
    "cci_params": {},
    "cluster_params": {},
    "psts_params": {},
    "dea_params": {},
    "lr_params": {},
}

# print(stlearn, file=sys.stdout)

app = Flask(__name__)
app.secret_key = b'_5#y2L"F4Q8z\n\xec]/'

UPLOAD_FOLDER = tempfile.mkdtemp()
print(UPLOAD_FOLDER)
TEMPLATES_AUTO_RELOAD = True
app.config["UPLOAD_FOLDER"] = UPLOAD_FOLDER
app.config["SEND_FILE_MAX_AGE_DEFAULT"] = 0
app.config["TEMPLATES_AUTO_RELOAD"] = TEMPLATES_AUTO_RELOAD
app.config["SESSION_PERMANENT"] = False


@app.route("/", methods=["GET"])
def index():
    return render_template("index.html", step_log=step_log)


@app.route("/upload")
def upload():
    return render_template("upload.html", step_log=step_log, flash_bool=True)


@app.route("/preprocessing", methods=["GET", "POST"])
def preprocessing():
    global adata, step_log
    updated_page = views.run_preprocessing(request, adata, step_log)
    return updated_page


@app.route("/clustering", methods=["GET", "POST"])
def clustering():
    global adata, step_log
    updated_page = views.run_clustering(request, adata, step_log)
    return updated_page


@app.route("/lr", methods=["GET", "POST"])
def lr():
    global adata, step_log
    updated_page = views.run_lr(request, adata, step_log)
    return updated_page


@app.route("/cci", methods=["GET", "POST"])
def cci():
    global adata, step_log
    updated_page = views.run_cci(request, adata, step_log)
    return updated_page


@app.route("/psts", methods=["GET", "POST"])
def psts():
    global adata, step_log

    if "clusters" not in adata.obs.columns:
        return redirect(url_for("choose_cluster"))
    else:
        updated_page = views.run_psts(request, adata, step_log)
        return updated_page


@app.route("/dea", methods=["GET", "POST"])
def dea():
    global adata, step_log
    updated_page = views.run_dea(request, adata, step_log)
    return updated_page


allow_files = [
    "filtered_feature_bc_matrix.h5",
    "tissue_hires_image.png",
    "tissue_lowres_image.png",
    "tissue_positions_list.csv",
    "scalefactors_json.json",
]


@app.route("/folder_uploader", methods=["GET", "POST"])
def folder_uploader():
    if request.method == "POST":
        # Clean uploads folder before upload a new data
        import shutil

        shutil.rmtree(app.config["UPLOAD_FOLDER"])
        os.makedirs(app.config["UPLOAD_FOLDER"])
        open(app.config["UPLOAD_FOLDER"] + "/.gitkeep", "a").close()
        # os.mknod()

        # Get list of files from selected folder
        files = request.files.getlist("file")

        os.mkdir(os.path.join(app.config["UPLOAD_FOLDER"], "spatial"))

        # allow_upload_files = list(map(lambda x: x ),allow_files)

        uploaded = []
        i = 0
        for file in files:

            filename = secure_filename(file.filename)

            if allow_files[0] in filename:
                file.save(os.path.join(app.config["UPLOAD_FOLDER"], filename))
                os.rename(
                    os.path.join(app.config["UPLOAD_FOLDER"], filename),
                    os.path.join(app.config["UPLOAD_FOLDER"], allow_files[0]),
                )
                uploaded.append(allow_files[0])

            for allow_file in allow_files[1:]:
                if allow_file in filename:
                    file.save(
                        os.path.join(app.config["UPLOAD_FOLDER"] + "/spatial", filename)
                    )
                    os.rename(
                        os.path.join(
                            app.config["UPLOAD_FOLDER"] + "/spatial", filename
                        ),
                        os.path.join(
                            app.config["UPLOAD_FOLDER"] + "/spatial", allow_file
                        ),
                    )

                    uploaded.append(allow_file)

            print(i)
            i += 1
            if len(uploaded) == 5:
                flash("File uploaded successfully")
                global adata, step_log
                # step_log = {
                #     "uploaded": [False, "Upload file"],
                #     "preprocessed": [False, "Preprocessing"],
                #     "clustering": [False, "Clustering"],
                #     "psts": [False, "Spatial trajectory"],
                #     "cci_rank": [False, "Cell-cell interaction"],
                #     "dea": [False, "Differential expression analysis"],
                #     # _params suffix important for templates/progress.html
                #     "preprocessed_params": {},
                #     "cci_params": {},
                #     "cluster_params": {},
                #     "psts_params": {},
                #     "dea_params": {},
                # }
                adata = stlearn.Read10X(app.config["UPLOAD_FOLDER"])
                adata.var_names_make_unique()  # removing duplicates
                # ensuring compatible format for CCI, since need _ to pair LRs #
                adata.var_names = numpy.array(
                    [var_name.replace("_", "-") for var_name in adata.var_names]
                )

                shutil.rmtree(app.config["UPLOAD_FOLDER"])

                step_log["uploaded"][0] = True

                return redirect(url_for("upload"))

        if len(uploaded) != 5:
            missing_files = []
            for file in allow_files:
                if file not in uploaded:
                    missing_files.append(file)
            shutil.rmtree(app.config["UPLOAD_FOLDER"])
            flash("Upload ERROR: Missing " + ", ".join(missing_files))
            return redirect(url_for("upload"))


@app.route("/file_uploader", methods=["GET", "POST"])
def file_uploader():
    if request.method == "POST":

        global adata, step_log

        # Clean uploads folder before upload a new data
        import shutil

        shutil.rmtree(app.config["UPLOAD_FOLDER"])
        os.makedirs(app.config["UPLOAD_FOLDER"])
        open(app.config["UPLOAD_FOLDER"] + "/.gitkeep", "a").close()
        # os.mknod()
        f = request.files["file"]
        filename = secure_filename(f.filename)
        f.save(os.path.join(app.config["UPLOAD_FOLDER"], filename))
        try:
            adata = scanpy.read_h5ad(app.config["UPLOAD_FOLDER"] + "/" + f.filename)
        except:
            flash("Upload ERROR: Please choose the right AnnData file ")

        ### Updating log file with current anndata state ###
        step_log["uploaded"][0] = True

        if "n_cells" in adata.var.columns:
            step_log["preprocessed"][0] = True

        for col in adata.obs.columns:
            if adata.obs[col].dtype.name == "category":
                if col != "sub_cluster_labels":
                    step_log["clustering"][0] = True

        if "global_graph" in adata.uns:
            step_log["psts"][0] = True

        step_log["lr"][0] = "lr_summary" in adata.uns
        step_log["cci"][0] = np.any(["lr_cci_" in key for key in adata.uns])

        return redirect(url_for("upload"))


@app.route("/choose_cluster", methods=["GET", "POST"])
def choose_cluster():
    menu = []

    for col in adata.obs.columns:
        if adata.obs[col].dtype.name == "category":
            if col != "sub_cluster_labels":
                menu.append(col)

    return render_template(
        "choose_cluster.html",
        template="Flask",
        relative_urls=False,
        step_log=step_log,
        menu=menu,
    )


@app.route("/convert_clusters", methods=["GET", "POST"])
def convert_clusters():
    if request.method == "POST":
        adata.obs["clusters"] = adata.obs[request.form["convert_clusters"]]
        scanpy.tl.paga(adata, groups="clusters")
        stlearn.pl.cluster_plot(adata, use_label="clusters")

        return redirect(url_for("psts"))


@app.route("/gene_plot")
def gene_plot():
    script = server_document("http://127.0.0.1:5006/bokeh_gene_plot")
    return render_template(
        "gene_plot.html",
        script=script,
        template="Flask",
        relative_urls=False,
        step_log=step_log,
    )


@app.route("/cluster_plot")
def cluster_plot():
    script = server_document("http://127.0.0.1:5006/bokeh_cluster_plot")
    return render_template(
        "cluster_plot.html",
        script=script,
        template="Flask",
        relative_urls=False,
        step_log=step_log,
    )


@app.route("/lr_plot")
def lr_plot():
    script = server_document("http://127.0.0.1:5006/bokeh_lr_plot")
    return render_template(
        "lr_plot.html",
        script=script,
        template="Flask",
        relative_urls=False,
        step_log=step_log,
    )


@app.route("/spatial_cci_plot")
def spatial_cci_plot():
    script = server_document("http://127.0.0.1:5006/bokeh_spatial_cci_plot")
    return render_template(
        "spatial_cci_plot.html",
        script=script,
        template="Flask",
        relative_urls=False,
        step_log=step_log,
    )


@app.route("/annotate_plot")
def annotate_plot():
    script = server_document("http://127.0.0.1:5006/bokeh_annotate_plot")
    return render_template(
        "annotate_plot.html",
        script=script,
        template="Flask",
        relative_urls=False,
        step_log=step_log,
    )


@app.route("/save_adata", methods=["POST"])
def save_adata():
    if request.method == "POST":
        fd, path = tempfile.mkstemp()
        from datetime import datetime

        now = datetime.now()
        date_time = now.strftime("%m-%d-%Y_%H-%M-%S")

        adata.write_h5ad(path)
        return send_file(
            path, as_attachment=True, attachment_filename="adata_" + date_time + ".h5ad"
        )


def modify_doc_gene_plot(doc):
    from stlearn.plotting.classes_bokeh import BokehGenePlot

    gp_object = BokehGenePlot(adata)
    doc.add_root(row(gp_object.layout, width=800))

    gp_object.data_alpha.on_change("value", gp_object.update_data)
    gp_object.tissue_alpha.on_change("value", gp_object.update_data)
    gp_object.spot_size.on_change("value", gp_object.update_data)
    gp_object.gene_select.on_change("value", gp_object.update_data)
    gp_object.cmap_select.on_change("value", gp_object.update_data)

    if len(gp_object.menu) != 0:
        gp_object.use_label.on_change("value", gp_object.update_data)
    gp_object.output_backend.on_change("value", gp_object.update_data)


def modify_doc_cluster_plot(doc):
    from stlearn.plotting.classes_bokeh import BokehClusterPlot

    gp_object = BokehClusterPlot(adata)
    doc.add_root(row(gp_object.layout, width=800))

    gp_object.use_label.on_change("value", gp_object.update_list)
    gp_object.use_label.on_change("value", gp_object.update_data)
    gp_object.data_alpha.on_change("value", gp_object.update_data)
    gp_object.tissue_alpha.on_change("value", gp_object.update_data)
    gp_object.spot_size.on_change("value", gp_object.update_data)
    gp_object.list_cluster.on_change("active", gp_object.update_data)
    gp_object.checkbox_group.on_change("active", gp_object.update_data)
    gp_object.output_backend.on_change("value", gp_object.update_data)
    if "rank_genes_groups" in adata.uns:
        gp_object.n_top_genes.on_change("value", gp_object.update_data)
        gp_object.cmap_select.on_change("value", gp_object.update_data)
        gp_object.plot_select.on_change("value", gp_object.update_data)
        gp_object.min_logfoldchange.on_change("value", gp_object.update_data)


def modify_doc_spatial_cci_plot(doc):
    from stlearn.plotting.classes_bokeh import BokehSpatialCciPlot

    gp_object = BokehSpatialCciPlot(adata)
    doc.add_root(row(gp_object.layout, width=800))

    gp_object.annot_select.on_change("value", gp_object.update_list)
    gp_object.annot_select.on_change("value", gp_object.update_data)
    gp_object.lr_select.on_change("value", gp_object.update_data)
    gp_object.data_alpha.on_change("value", gp_object.update_data)
    gp_object.tissue_alpha.on_change("value", gp_object.update_data)
    gp_object.spot_size.on_change("value", gp_object.update_data)
    gp_object.list_cluster.on_change("active", gp_object.update_data)
    gp_object.output_backend.on_change("value", gp_object.update_data)


def modify_doc_lr_plot(doc):
    from stlearn.plotting.classes_bokeh import BokehLRPlot

    gp_object = BokehLRPlot(adata)
    doc.add_root(row(gp_object.layout, width=800))

    gp_object.data_alpha.on_change("value", gp_object.update_data)
    gp_object.tissue_alpha.on_change("value", gp_object.update_data)
    gp_object.spot_size.on_change("value", gp_object.update_data)
    # gp_object.het_select.on_change("value", gp_object.update_data)
    gp_object.lr_select.on_change("value", gp_object.update_data)
    gp_object.output_backend.on_change("value", gp_object.update_data)


def modify_doc_annotate_plot(doc):
    from stlearn.plotting.classes_bokeh import Annotate

    gp_object = Annotate(adata)
    doc.add_root(row(gp_object.layout, width=800))
    gp_object.data_alpha.on_change("value", gp_object.update_data)
    gp_object.tissue_alpha.on_change("value", gp_object.update_data)
    gp_object.spot_size.on_change("value", gp_object.update_data)


# App for gene_plot
bkapp = Application(FunctionHandler(modify_doc_gene_plot))

# App for cluster_plot
bkapp2 = Application(FunctionHandler(modify_doc_cluster_plot))

# App for lr_plot
bkapp3 = Application(FunctionHandler(modify_doc_lr_plot))

# App for cci_spatial_plot
bkapp3_1 = Application(FunctionHandler(modify_doc_spatial_cci_plot))

# App for annotate_plot
bkapp4 = Application(FunctionHandler(modify_doc_annotate_plot))


def bk_worker():
    asyncio.set_event_loop(asyncio.new_event_loop())

    server = Server(
        {
            "/bokeh_gene_plot": bkapp,
            "/bokeh_cluster_plot": bkapp2,
            # "/bokeh_cci_plot": bkapp3,
            "/bokeh_lr_plot": bkapp3,
            "/bokeh_spatial_cci_plot": bkapp3_1,
            "/bokeh_annotate_plot": bkapp4,
        },
        io_loop=IOLoop(),
        allow_websocket_origin=["127.0.0.1:5000", "localhost:5000"],
    )
    server.start()
    server.io_loop.start()


from threading import Thread

Thread(target=bk_worker).start()



import numpy as np


def compare_transitions(adata, trajectories):
    """\
    Compare transition markers between two clades

    Parameters
    ----------
    adata
        Annotated data matrix.
    trajectories
        List of clades names user want to compare.
    Returns
    -------
    Anndata
    """

    pos_1 = list(
        adata.uns[trajectories[0]][adata.uns[trajectories[0]]["score"] >= 0]["gene"]
    )
    pos_2 = list(
        adata.uns[trajectories[1]][adata.uns[trajectories[1]]["score"] >= 0]["gene"]
    )
    compare_pos_1 = np.setdiff1d(pos_1, pos_2, assume_unique=True)
    compare_pos_2 = np.setdiff1d(pos_2, pos_1, assume_unique=True)

    neg_1 = list(
        adata.uns[trajectories[0]][adata.uns[trajectories[0]]["score"] < 0]["gene"]
    )
    neg_2 = list(
        adata.uns[trajectories[1]][adata.uns[trajectories[1]]["score"] < 0]["gene"]
    )
    compare_neg_1 = np.setdiff1d(neg_1, neg_2, assume_unique=True)
    compare_neg_2 = np.setdiff1d(neg_2, neg_1, assume_unique=True)

    compare_result = {}
    compare_result["pos_1"] = compare_pos_1
    compare_result["pos_2"] = compare_pos_2
    compare_result["neg_1"] = compare_neg_1
    compare_result["neg_2"] = compare_neg_2

    compare_result["trajectories"] = trajectories

    adata.uns["compare_result"] = compare_result
    print(
        "The result of comparison between "
        + trajectories[0]
        + " and "
        + trajectories[1]
        + " stored in 'adata.uns['compare_result']'"
    )



from typing import Optional, Union
from anndata import AnnData
from pathlib import Path
import os
import pandas as pd
import numpy as np
from natsort import natsorted


def labels(
    adata: AnnData,
    label_filepath: str = None,
    index_col: int = 0,
    use_label: str = None,
    sep: str = "\t",
    copy: bool = False,
) -> Optional[AnnData]:
    """\
    Add label transfer results into AnnData object

    Parameters
    ----------
    adata: AnnData          The data object to add L-R info into
    label_filepath: str     The path to the label transfer results file
    use_label: str          Where to store the label_transfer results, defaults to 'predictions' in adata.obs & 'label_transfer' in adata.uns.
    sep: str                Separator of the csv file
    copy: bool              Copy flag indicating copy or direct edit

    Returns
    -------
    adata: AnnData          The data object that L-R added into

    """
    labels = pd.read_csv(label_filepath, index_col=index_col, sep=sep)
    uns_key = "label_transfer" if type(use_label) == type(None) else use_label
    adata.uns[uns_key] = labels.drop(["predicted.id", "prediction.score.max"], axis=1)

    key_add = "predictions" if type(use_label) == type(None) else use_label
    key_source = "predicted.id"
    adata.obs[key_add] = pd.Categorical(
        values=np.array(labels[key_source]).astype("U"),
        categories=natsorted(labels[key_source].unique().astype("U")),
    )

    print(f"label transfer results added to adata.uns['{uns_key}']")
    print(f"predicted label added to adata.obs['{key_add}'].")

    return adata if copy else None



""" This is more a general views focussed on defining functions which are \
	called by other views for specify pages. This way different pages can be \
	used to display different data, but in a consistent way.
"""

import sys
import numpy
import numpy as np
from flask import flash
from source.forms import forms

from source.forms.utils import flash_errors
import source.forms.view_helpers as vhs
import traceback

from flask import render_template

import scanpy as sc
import stlearn as st

from scipy.spatial.distance import cosine

# Creating the forms using a class generator #
PreprocessForm = forms.getPreprocessForm()
# CCIForm = forms.getCCIForm() #OLD
ClusterForm = forms.getClusterForm()
LRForm = forms.getLRForm()


def run_preprocessing(request, adata, step_log):
    """Performs the scanpy pre-processing steps based on the inputted data."""

    form = PreprocessForm(request.form)

    if not form.validate_on_submit():
        flash_errors(form)

    elif type(adata) == type(None):
        flash("Need to load data first!")

    else:
        # Logging params used #
        step_log["preprocessed_params"] = vhs.getData(form)
        print(step_log["preprocessed_params"], file=sys.stdout)

        # QC filtering #
        sc.pp.filter_cells(adata, min_genes=vhs.getVal(form, "Minimum genes per spot"))
        sc.pp.filter_cells(
            adata, min_counts=vhs.getVal(form, "Minimum counts per spot")
        )
        sc.pp.filter_genes(adata, min_cells=vhs.getVal(form, "Minimum spots per gene"))
        sc.pp.filter_genes(
            adata, min_counts=vhs.getVal(form, "Minimum counts per gene")
        )

        # Pre-processing #
        if vhs.getVal(form, "Normalize total"):
            sc.pp.normalize_total(adata, target_sum=1e4)
        if vhs.getVal(form, "Log 1P"):
            sc.pp.log1p(adata)
        adata.raw = adata
        if vhs.getVal(form, "Scale"):
            sc.pp.scale(adata, max_value=10)

        # Setting pre-process to true #
        step_log["preprocessed"][0] = True

    if step_log["preprocessed"][0]:
        flash("Preprocessing is completed!")

    updated_page = render_template(
        "preprocessing.html",
        title=step_log["preprocessed"][1],
        preprocess_form=form,
        flash_bool=step_log["preprocessed"][0],
        step_log=step_log,
    )

    return updated_page


def run_lr(request, adata, step_log):
    """Runs LR analysis."""

    form = LRForm(request.form)

    if not form.validate_on_submit():
        flash_errors(form)

    elif type(adata) == type(None):
        flash("Need to load data first!")

    else:
        step_log["lr_params"] = vhs.getData(form)
        print(step_log["lr_params"], file=sys.stdout)
        elements = numpy.array(list(step_log["lr_params"].keys()))
        # order: Species, Spot neighbourhood, min_spots, n_pairs, CPUs
        element_values = list(step_log["lr_params"].values())
        dist = element_values[1]
        dist = dist if dist != -1 else None

        # Loading the LR databases available within stlearn (from NATMI)
        lrs = st.tl.cci.load_lrs(["connectomeDB2020_lit"], species=element_values[0])

        # Running the analysis #
        st.tl.cci.run(
            adata,
            lrs,
            min_spots=element_values[2],
            distance=dist,
            n_pairs=element_values[3],
            n_cpus=element_values[-1],
        )
        flash("LR analysis is completed!")

    step_log["lr"][0] = "lr_summary" in adata.uns

    updated_page = render_template(
        "lr.html",
        title=step_log["lr"][1],
        lr_form=form,
        flash_bool=True,
        step_log=step_log,
    )
    return updated_page


def run_cci(request, adata, step_log):
    """Performs CCI analysis."""

    CCIForm = forms.getCCIForm(adata)
    form = CCIForm(request.form)

    if not form.validate_on_submit():
        flash_errors(form)

    elif type(adata) == type(None):
        flash("Need to load data first!")

    else:
        step_log["cci_params"] = vhs.getData(form)
        print(step_log["cci_params"], file=sys.stdout)
        elements = numpy.array(list(step_log["cci_params"].keys()))
        # order: cell_type, min_spots, spot_mixtures, cell_prop_cutoff, sig_spots
        #           n_perms
        element_values = list(step_log["cci_params"].values())

        if not form.validate_on_submit():
            flash_errors(form)

        else:
            try:
                # Running the counting of co-occurence of cell types and LR expression #
                st.tl.cci.run_cci(
                    adata,
                    element_values[0],
                    min_spots=element_values[1],
                    spot_mixtures=element_values[2],
                    cell_prop_cutoff=element_values[3],
                    sig_spots=True,  # Should make this not optional..
                    n_perms=element_values[4],
                )

                flash("CCI analysis is completed!")

            except Exception as msg:
                traceback.print_exc(file=sys.stdout)
                flash("Analysis ERROR: " + str(msg))
                print(msg)

    step_log["cci"][0] = np.any(["lr_cci_" in key for key in adata.uns])

    updated_page = render_template(
        "cci.html",
        title=step_log["cci"][1],
        cci_form=form,
        flash_bool=True,
        step_log=step_log,
    )

    return updated_page


def run_clustering(request, adata, step_log):
    """Performs clustering analysis."""

    form = ClusterForm(request.form)

    step_log["cluster_params"] = vhs.getData(form)
    print(step_log["cluster_params"], file=sys.stdout)
    elements = list(step_log["cluster_params"].keys())
    # order: pca_comps, SME bool, method, method_param
    element_values = list(step_log["cluster_params"].values())

    if not form.validate_on_submit():
        flash_errors(form)

    elif type(adata) == type(None):
        flash("Need to load data first!")

    else:
        try:
            # Running PCA, performs scaling internally #
            n_comps = element_values[0]
            st.em.run_pca(adata, n_comps=n_comps)

            print(element_values[1], file=sys.stdout, flush=True)
            if element_values[1]:  # Performing SME clustering #
                # Image feature extraction #
                st.pp.tiling(adata)
                st.pp.extract_feature(adata)

                # apply stSME to data (format of data depending on preprocess)
                st.spatial.SME.SME_normalize(adata, use_data="raw")
                adata.X = adata.obsm["raw_SME_normalized"]
                st.em.run_pca(adata, n_comps=n_comps)

            # Performing the clustering on the PCA #
            if element_values[2] == "KMeans":  # KMeans
                param = int(element_values[3])
                st.tl.clustering.kmeans(adata, n_clusters=param, use_data="X_pca")

                st.pp.neighbors(adata, n_neighbors=element_values[5], use_rep="X_pca")
                sc.tl.paga(adata, groups="kmeans")
                st.pl.cluster_plot(adata, use_label="kmeans")

            elif element_values[2] == "Louvain":  # Louvain
                param = element_values[4]
                st.pp.neighbors(adata, n_neighbors=element_values[5], use_rep="X_pca")
                st.tl.clustering.louvain(adata, resolution=param)
                sc.tl.paga(adata, groups="louvain")
                st.pl.cluster_plot(adata, use_label="louvain")

            else:  # Leiden
                param = element_values[4]
                st.pp.neighbors(adata, n_neighbors=element_values[5], use_rep="X_pca")
                sc.tl.leiden(adata, resolution=param)
                sc.tl.paga(adata, groups="leiden")
                st.pl.cluster_plot(adata, use_label="leiden")

            step_log["clustering"][0] = True
            flash("Clustering is completed!")

        except Exception as msg:
            traceback.print_exc(file=sys.stdout)
            flash("Analysis ERROR: " + str(msg))
            print(msg)

    updated_page = render_template(
        "clustering.html",
        title=step_log["clustering"][1],
        clustering_form=form,
        flash_bool=True,
        step_log=step_log,
    )

    return updated_page


def run_psts(request, adata, step_log):
    """Performs psts analysis; must have performed clustering first."""
    # Creating the form with the clustering information #
    cluster_set = numpy.unique(adata.obs["clusters"].values)
    order = numpy.argsort([int(cluster) for cluster in cluster_set])
    cluster_set = cluster_set[order]

    options = ["Auto", "Spatial distance only", "Gene expression distance only"]

    from .utils import get_all_paths

    trajectory_set = get_all_paths(adata)

    PSTSForm = forms.getPSTSForm(trajectory_set, cluster_set, options)
    form = PSTSForm(request.form)

    step_log["psts_params"] = vhs.getData(form)
    print(step_log["psts_params"], file=sys.stdout)
    elements = list(step_log["psts_params"].keys())
    # order: pca_comps, SME bool, method, method_param
    element_values = list(step_log["psts_params"].values())

    if element_values[4] == "Auto":
        w = None
    elif element_values[4] == "Spatial distance only":
        w = 0
    else:
        w = 1

    if not form.validate_on_submit():
        flash_errors(form)

    elif type(adata) == type(None):
        flash("Need to load data first!")

    else:
        try:
            from stlearn.spatials.trajectory import set_root

            root_index = set_root(
                adata, use_label="clusters", cluster=str(element_values[0])
            )

            adata.uns["iroot"] = root_index

            print(root_index, file=sys.stdout, flush=True)

            # Performing the TI #
            print(element_values[3], file=sys.stdout, flush=True)

            node_order = element_values[3].split(" - ")

            st.spatial.trajectory.pseudotime(
                adata,
                eps=element_values[2],
                use_rep="X_pca",
                use_label="clusters",
                reverse=element_values[1],
            )
            print(node_order)
            st.spatial.trajectory.pseudotimespace_global(
                adata, use_label="clusters", list_clusters=node_order, w=w
            )

            st.pl.cluster_plot(
                adata,
                use_label="clusters",
                show_trajectories=True,
                list_clusters=node_order,
                show_subcluster=True,
            )

            step_log["psts"][0] = True
            flash("Trajectory inference is completed!")

        except Exception as msg:
            traceback.print_exc(file=sys.stdout)
            flash("Analysis ERROR: " + str(msg))
            print(msg)

    updated_page = render_template(
        "psts.html",
        title=step_log["psts"][1],
        psts_form=form,
        flash_bool=True,
        step_log=step_log,
    )

    return updated_page


def run_dea(request, adata, step_log):

    list_labels = []

    for col in adata.obs.columns:
        if adata.obs[col].dtype.name == "category":
            if col != "sub_cluster_labels":
                list_labels.append(col)

    list_labels = numpy.array(list_labels)

    methods = numpy.array(["t-test", "t-test_overestim_var", "logreg", "wilcoxon"])

    DEAForm = forms.getDEAForm(list_labels, methods)
    form = DEAForm(request.form)

    step_log["dea_params"] = vhs.getData(form)
    print(step_log["dea_params"], file=sys.stdout)
    elements = list(step_log["dea_params"].keys())
    element_values = list(step_log["dea_params"].values())

    if not form.validate_on_submit():
        flash_errors(form)

    elif type(adata) == type(None):
        flash("Need to load data first!")

    else:
        try:

            sc.tl.rank_genes_groups(adata, element_values[0], method=element_values[1])

            step_log["dea"][0] = True
            flash("Differential expression analysis is completed!")

        except Exception as msg:
            traceback.print_exc(file=sys.stdout)
            flash("Analysis ERROR: " + str(msg))
            print(msg)

    updated_page = render_template(
        "dea.html",
        title=step_log["dea"][1],
        dea_form=form,
        flash_bool=True,
        step_log=step_log,
    )

    return updated_page



from typing import Optional, Union
from anndata import AnnData
from matplotlib import pyplot as plt
from pathlib import Path
import os
from PIL import Image

Image.MAX_IMAGE_PIXELS = None


def image(
    adata: AnnData,
    imgpath: Union[Path, str],
    library_id: str,
    quality: str = "hires",
    scale: float = 1.0,
    visium: bool = False,
    spot_diameter_fullres: float = 50,
    copy: bool = False,
) -> Optional[AnnData]:

    """\
    Adding image data to the Anndata object

    Parameters
    ----------
    adata
        Annotated data matrix.
    imgpath
        Image path.
    library_id
        Identifier for the visium library. Can be modified when concatenating multiple adata objects.
    scale
        Set scale factor.
    quality
        Set quality that convert to stlearn to use. Store in anndata.obs['imagecol' & 'imagerow'].
    visium
        Is this anndata read from Visium platform or not.
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    Depending on `copy`, returns or updates `adata` with the following fields.
    **tissue_img** : `adata.uns` field
        Array format of image, saving by Pillow package.
    """

    if imgpath is not None and os.path.isfile(imgpath):
        try:
            img = plt.imread(imgpath, 0)

            if visium:
                adata.uns["spatial"][library_id]["images"][quality] = img
            else:
                adata.uns["spatial"] = {}
                adata.uns["spatial"][library_id] = {}
                adata.uns["spatial"][library_id]["images"] = {}
                adata.uns["spatial"][library_id]["images"][quality] = img
                adata.uns["spatial"][library_id]["use_quality"] = quality
                adata.uns["spatial"][library_id]["scalefactors"] = {}
                adata.uns["spatial"][library_id]["scalefactors"][
                    "tissue_" + quality + "_scalef"
                ] = scale
                adata.uns["spatial"][library_id]["scalefactors"][
                    "spot_diameter_fullres"
                ] = spot_diameter_fullres
                adata.obsm["spatial"] = adata.obs[["imagecol", "imagerow"]].values
                adata.obs[["imagecol", "imagerow"]] = adata.obsm["spatial"] * scale

            print("Added tissue image to the object!")

            return adata if copy else None
        except:
            raise ValueError(
                f"""\
            {imgpath!r} does not end on a valid extension.
            """
            )
    else:
        raise ValueError(
            f"""\
        {imgpath!r} does not end on a valid extension.
        """
        )
    return adata if copy else None



from anndata import AnnData
from typing import Optional, Union
import numpy as np
from stlearn.spatials.trajectory.utils import _correlation_test_helper


def set_root(adata: AnnData, use_label: str, cluster: str, use_raw: bool = False):

    """\
    Automatically set the root index.

    Parameters
    ----------
    adata
        Annotated data matrix.
    use_label
        Use label result of cluster method.
    cluster
        Choose cluster to use as root
    use_raw
        Use the raw layer
    Returns
    -------
    Root index
    """

    tmp_adata = adata.copy()

    # Subset the data based on the chosen cluster

    tmp_adata = tmp_adata[
        tmp_adata.obs[tmp_adata.obs[use_label] == str(cluster)].index, :
    ]
    if use_raw == True:
        tmp_adata = tmp_adata.raw.to_adata()

    # Borrow from Cellrank to calculate CytoTrace score
    num_exp_genes = np.array((tmp_adata.X > 0).sum(axis=1)).reshape(-1)
    gene_corr, _, _, _ = _correlation_test_helper(tmp_adata.X.T, num_exp_genes[:, None])
    tmp_adata.var["gene_corr"] = gene_corr

    # Use top 1000 genes rather than top 200 genes
    top_1000 = tmp_adata.var.sort_values(by="gene_corr", ascending=False).index[:1000]
    tmp_adata.var["correlates"] = False
    tmp_adata.var.loc[top_1000, "correlates"] = True
    corr_mask = tmp_adata.var["correlates"]
    imputed_exp = tmp_adata[:, corr_mask].X

    # Scale ct score
    cytotrace_score = np.mean(imputed_exp, axis=1)
    cytotrace_score -= np.min(cytotrace_score)
    cytotrace_score /= np.max(cytotrace_score)

    # Get the root index
    local_index = np.argmax(cytotrace_score)
    obs_name = tmp_adata.obs.iloc[local_index].name

    return np.where(adata.obs_names == obs_name)[0][0]



from typing import Optional, Union
from anndata import AnnData
from pathlib import Path
import numpy as np
from scipy.sparse import csr_matrix
import pandas as pd
from ._weighting_matrix import (
    calculate_weight_matrix,
    impute_neighbour,
    _WEIGHTING_MATRIX,
    _PLATFORM,
)
import stlearn
import scipy
from ..._compat import Literal


def SME_impute0(
    adata: AnnData,
    use_data: str = "raw",
    weights: _WEIGHTING_MATRIX = "weights_matrix_all",
    platform: _PLATFORM = "Visium",
    copy: bool = False,
) -> Optional[AnnData]:
    """\
    using spatial location (S), tissue morphological feature (M) and gene expression (E) information to impute missing
    values

    Parameters
    ----------
    adata
        Annotated data matrix.
    use_data
        input data, can be `raw` counts or log transformed data
    weights
        weighting matrix for imputation.
        if `weights_matrix_all`, matrix combined all information from spatial location (S),
        tissue morphological feature (M) and gene expression (E)
        if `weights_matrix_pd_md`, matrix combined information from spatial location (S),
        tissue morphological feature (M)
    platform
        `Visium` or `Old_ST`
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    Anndata
    """
    if use_data == "raw":
        if isinstance(adata.X, csr_matrix):
            count_embed = adata.X.toarray()
        elif isinstance(adata.X, np.ndarray):
            count_embed = adata.X
        elif isinstance(adata.X, pd.Dataframe):
            count_embed = adata.X.values
        else:
            raise ValueError(
                f"""\
                    {type(adata.X)} is not a valid type.
                    """
            )
    else:
        count_embed = adata.obsm[use_data]

    calculate_weight_matrix(adata, platform=platform)

    impute_neighbour(adata, count_embed=count_embed, weights=weights)

    imputed_data = adata.obsm["imputed_data"].astype(float)
    mask = count_embed != 0
    count_embed_ = count_embed.astype(float)
    count_embed_[count_embed_ == 0] = np.nan
    adjusted_count_matrix = np.nanmean(np.array([count_embed_, imputed_data]), axis=0)
    adjusted_count_matrix[mask] = count_embed[mask]

    key_added = use_data + "_SME_imputed"
    adata.obsm[key_added] = adjusted_count_matrix

    print("The data adjusted by SME is added to adata.obsm['" + key_added + "']")

    return adata if copy else None


_COPY = Literal["pseudo_spot_adata", "combined_adata"]


def pseudo_spot(
    adata: AnnData,
    tile_path: Union[Path, str] = Path("/tmp/tiles"),
    use_data: str = "raw",
    crop_size: int = "auto",
    platform: _PLATFORM = "Visium",
    weights: _WEIGHTING_MATRIX = "weights_matrix_all",
    copy: _COPY = "pseudo_spot_adata",
) -> Optional[AnnData]:
    """\
    using spatial location (S), tissue morphological feature (M) and gene expression (E) information to impute
    gap between spots and increase resolution for gene detection

    Parameters
    ----------
    adata
        Annotated data matrix.
    use_data
        Input data, can be `raw` counts, log transformed data or dimension reduced space(`X_pca` and `X_umap`)
    tile_path
        Path to save spot image tiles
    crop_size
        Size of tiles
        if `auto`, automatically detect crop size
    weights
        Weighting matrix for imputation.
        if `weights_matrix_all`, matrix combined all information from spatial location (S),
        tissue morphological feature (M) and gene expression (E)
        if `weights_matrix_pd_md`, matrix combined information from spatial location (S),
        tissue morphological feature (M)
    platform
        `Visium` or `Old_ST`
    copy
        Return Anndata
        if `pseudo_spot_adata`, imputed Anndata
        if `combined_adata`, merged Anndata of original data imputed Anndata.
    Returns
    -------
    Anndata
    """
    from sklearn.linear_model import LinearRegression
    import math

    if platform == "Visium":
        img_row = adata.obs["imagerow"]
        img_col = adata.obs["imagecol"]
        array_row = adata.obs["array_row"]
        array_col = adata.obs["array_col"]
        rate = 3
        obs_df_ = adata.obs[["array_row", "array_col"]].copy()
        obs_df_.loc[:, "array_row"] = obs_df_["array_row"].apply(lambda x: x - 2 / 3)
        obs_df = adata.obs[["array_row", "array_col"]].copy()
        obs_df.loc[:, "array_row"] = obs_df["array_row"].apply(lambda x: x + 2 / 3)
        obs_df = obs_df.append(obs_df_).reset_index()
        obs_df.drop_duplicates(subset=["array_row", "array_col"], keep="last")

    elif platform == "Old_ST":
        img_row = adata.obs["imagerow"]
        img_col = adata.obs["imagecol"]
        array_row = adata.obs_names.map(lambda x: x.split("x")[1])
        array_col = adata.obs_names.map(lambda x: x.split("x")[0])
        rate = 1.5
        obs_df_left = pd.DataFrame(
            {"array_row": array_row.to_list(), "array_col": array_col.to_list()},
            dtype=np.float64,
        )
        obs_df_left.loc[:, "array_row"] = obs_df_left["array_row"].apply(
            lambda x: x - 1 / 2
        )

        obs_df_right = pd.DataFrame(
            {"array_row": array_row.to_list(), "array_col": array_col.to_list()},
            dtype=np.float64,
        )
        obs_df_right.loc[:, "array_row"] = obs_df_right["array_row"].apply(
            lambda x: x + 1 / 2
        )

        obs_df_up = pd.DataFrame(
            {"array_row": array_row.to_list(), "array_col": array_col.to_list()},
            dtype=np.float64,
        )
        obs_df_up.loc[:, "array_col"] = obs_df_up["array_col"].apply(
            lambda x: x - 1 / 2
        )

        obs_df_down = pd.DataFrame(
            {"array_row": array_row.to_list(), "array_col": array_col.to_list()},
            dtype=np.float64,
        )
        obs_df_down.loc[:, "array_col"] = obs_df_down["array_col"].apply(
            lambda x: x + 1 / 2
        )

        obs_df_left_up = pd.DataFrame(
            {"array_row": array_row.to_list(), "array_col": array_col.to_list()},
            dtype=np.float64,
        )
        obs_df_left_up.loc[:, "array_row"] = obs_df_left_up["array_row"].apply(
            lambda x: x - 1 / 2
        )
        obs_df_left_up.loc[:, "array_col"] = obs_df_left_up["array_col"].apply(
            lambda x: x - 1 / 2
        )

        obs_df_right_up = pd.DataFrame(
            {"array_row": array_row.to_list(), "array_col": array_col.to_list()},
            dtype=np.float64,
        )
        obs_df_right_up.loc[:, "array_row"] = obs_df_right_up["array_row"].apply(
            lambda x: x + 1 / 2
        )
        obs_df_right_up.loc[:, "array_col"] = obs_df_right_up["array_col"].apply(
            lambda x: x - 1 / 2
        )

        obs_df_left_down = pd.DataFrame(
            {"array_row": array_row.to_list(), "array_col": array_col.to_list()},
            dtype=np.float64,
        )
        obs_df_left_down.loc[:, "array_row"] = obs_df_left_down["array_row"].apply(
            lambda x: x - 1 / 2
        )
        obs_df_left_down.loc[:, "array_col"] = obs_df_left_down["array_col"].apply(
            lambda x: x + 1 / 2
        )

        obs_df_right_down = pd.DataFrame(
            {"array_row": array_row.to_list(), "array_col": array_col.to_list()},
            dtype=np.float64,
        )
        obs_df_right_down.loc[:, "array_row"] = obs_df_right_down["array_row"].apply(
            lambda x: x + 1 / 2
        )
        obs_df_right_down.loc[:, "array_col"] = obs_df_right_down["array_col"].apply(
            lambda x: x + 1 / 2
        )

        obs_df = obs_df_left.append(
            [
                obs_df_right,
                obs_df_up,
                obs_df_down,
                obs_df_left_up,
                obs_df_right_up,
                obs_df_left_down,
                obs_df_right_down,
            ]
        ).reset_index()
        obs_df.drop_duplicates(subset=["array_row", "array_col"], keep="last")
    else:
        raise ValueError(
            f"""\
                {platform!r} does not support.
                """
        )

    reg_row = LinearRegression().fit(array_row.values.reshape(-1, 1), img_row)

    reg_col = LinearRegression().fit(array_col.values.reshape(-1, 1), img_col)

    obs_df.loc[:, "imagerow"] = (
        obs_df.loc[:, "array_row"] * reg_row.coef_ + reg_row.intercept_
    )
    obs_df.loc[:, "imagecol"] = (
        obs_df.loc[:, "array_col"] * reg_col.coef_ + reg_col.intercept_
    )

    impute_coor = obs_df[["imagecol", "imagerow"]]
    coor = adata.obs[["imagecol", "imagerow"]].append(impute_coor)

    point_tree = scipy.spatial.cKDTree(coor)
    n_neighbour = []
    unit = math.sqrt(reg_row.coef_**2 + reg_col.coef_**2)
    for i in range(len(impute_coor)):
        current_neighbour = point_tree.query_ball_point(
            impute_coor.values[i], round(unit)
        )

        current_neighbour = [x for x in current_neighbour if x < len(adata)]
        n_neighbour.append(len(current_neighbour))

    obs_df["n_neighbour"] = n_neighbour
    obs_df = obs_df.loc[obs_df["n_neighbour"] > 1, :].reset_index()

    obs_df.index = obs_df.index.map(lambda x: "Pseudo_Spot_" + str(x))

    impute_df = pd.DataFrame(0, index=obs_df.index, columns=adata.var_names)

    pseudo_spot_adata = AnnData(impute_df, obs=obs_df)
    pseudo_spot_adata.uns["spatial"] = adata.uns["spatial"]

    if crop_size == "auto":
        crop_size = round(unit / 2)

    stlearn.pp.tiling(pseudo_spot_adata, tile_path, crop_size=crop_size)

    stlearn.pp.extract_feature(pseudo_spot_adata)

    if use_data == "raw":
        if isinstance(adata.X, csr_matrix):
            count_embed = adata.X.toarray()
        elif isinstance(adata.X, np.ndarray):
            count_embed = adata.X
        elif isinstance(adata.X, pd.Dataframe):
            count_embed = adata.X.values
        else:
            print(f"{type(adata.X)} is not a valid type")
    else:
        count_embed = adata.obsm[use_data]

    calculate_weight_matrix(
        adata, pseudo_spot_adata, pseudo_spots=True, platform=platform
    )

    impute_neighbour(pseudo_spot_adata, count_embed=count_embed, weights=weights)

    assert pseudo_spot_adata.shape == pseudo_spot_adata.obsm["imputed_data"].shape

    pseudo_spot_adata.X = pseudo_spot_adata.obsm["imputed_data"]

    pseudo_spot_adata = pseudo_spot_adata[np.sum(pseudo_spot_adata.X, axis=1) > 0]

    print("Done")

    if copy == "pseudo_spot_adata":
        return pseudo_spot_adata
    else:
        return _merge(adata, pseudo_spot_adata)


def _merge(
    adata1: AnnData,
    adata2: AnnData,
    copy: bool = True,
) -> Optional[AnnData]:
    merged_df = adata1.to_df().append(adata2.to_df())
    merged_df_obs = adata1.obs.append(adata2.obs)
    merged_adata = AnnData(merged_df, obs=merged_df_obs)
    merged_adata.uns["spatial"] = adata1.uns["spatial"]

    return merged_adata if copy else None



from .adds.add_image import image
from .adds.add_positions import positions
from .adds.parsing import parsing
from .adds.add_lr import lr
from .adds.annotation import annotation
from .adds.add_labels import labels
from .adds.add_deconvolution import add_deconvolution
from .adds.add_mask import add_mask
from .adds.add_mask import apply_mask
from .adds.add_loupe_clusters import add_loupe_clusters



"""
Title: SpatialBasePlot for all spatial coordinates and image plot
Author: Duy Pham
Date: 20 Feb 2021
"""

from typing import Optional, Union, Mapping  # Special
from typing import Sequence, Iterable  # ABCs
from typing import Tuple  # Classes

import numpy as np
from anndata import AnnData

from .utils import (
    Empty,
    _empty,
    _check_spatial_data,
    _check_img,
    _check_spot_size,
    _check_scale_factor,
    _check_coords,
)


class Spatial(object):
    def __init__(
        self,
        adata: AnnData,
        basis: str = "spatial",
        img: Union[np.ndarray, None] = None,
        img_key: Union[str, None, Empty] = _empty,
        library_id: Union[str, None] = _empty,
        crop_coord: Optional[bool] = True,
        bw: Optional[bool] = False,
        scale_factor: Optional[float] = None,
        spot_size: Optional[float] = None,
        use_raw: Optional[bool] = False,
        **kwargs,
    ):

        self.adata = (adata,)
        self.library_id, self.spatial_data = _check_spatial_data(adata.uns, library_id)
        self.img, self.img_key = _check_img(self.spatial_data, img, img_key, bw=bw)
        self.spot_size = _check_spot_size(self.spatial_data, spot_size)
        self.scale_factor = _check_scale_factor(
            self.spatial_data, img_key=self.img_key, scale_factor=scale_factor
        )
        self.crop_coord = crop_coord
        self.use_raw = use_raw
        self.imagecol, self.imagerow = _check_coords(adata.obsm, self.scale_factor)



from anndata import AnnData
from stlearn.plotting.classes_bokeh import Annotate
from bokeh.io import output_notebook
from bokeh.plotting import show


def annotate_interactive(
    adata: AnnData,
):
    """\
    Allow user to manually define the clusters

    Parameters
    -------------------------------------
    adata
        Annotated data matrix.
    """

    bokeh_object = Annotate(adata)
    output_notebook()
    show(bokeh_object.app, notebook_handle=True)
